import os
import numpy as np
#from numba import njit
import matplotlib.pyplot as plt

import timeit as time
class Timer():
    def __init__(self):
        self.tstart = time.default_timer()
    def stop(self):
        dt = time.default_timer() - self.tstart
        print('Elapsed: %f' % dt)
        return dt

#@njit(cache=True)
def isInBox(b, x, y, z=None):
    """Find whether set of points of coords x, y
    is in the box b = [[xmin, ymin, zmin],
                       [xmax, ymax, zmax]]"""
    if len(b) != 2:
        print("the box not correct")
    e = 1e-6 * np.max(np.abs(b.ravel())) + 1e-6 * np.std(b.ravel())
    if z is None:
        return (
            ((b[0, 0] - e) < x)
            * ((b[0, 1] - e) < y)
            * (x < (b[1, 0] + e))
            * (y < (b[1, 1] + e))
        )
    else:
        return (
            ((b[0, 0] - e) < x)
            * ((b[0, 1] - e) < y)
            * ((b[0, 2] - e) < z)
            * (x < (b[1, 0] + e))
            * (y < (b[1, 1] + e))
            * (z < (b[1, 2] + e))
        )


#@njit(cache=True)
def meshgrid(a, b):
    A = a.repeat(len(b)).reshape((-1, len(b))).T
    B = b.repeat(len(a)).reshape((-1, len(a)))
    return A, B

def PlotMeshImage(f, m, cam, U=None, plot='mesh', newfig=True):
    """Plotting the mesh over the image. 

    Parameters
    ----------
    f : pyxel.Image
        The image over which to plot the mesh
    m : pyxel.Mesh
        The mesh
    cam : pyxel.Camera
        The camera model
    U : Numpy array
        A displacement dof vector (OPTIONNAL) to warp the mesh.
    plot : String (OPTIONNAL)
        'mesh': plot the mesh in yellow (DEFAULT)
        'displ': plot contour displacement field
        'strain': plot contour strain field

    """
    n = m.n.copy()
    if U is not None:
        n += U[m.conn]
    if newfig:
        plt.figure()
    f.Plot()
    u, v = cam.P(n[:, 0], n[:, 1])
    if plot == 'mesh':
        m.Plot(n=np.c_[v, u], edgecolor="y", alpha=0.6)
    elif plot == 'strain':
        m.PlotContourStrain(U, n=np.c_[v, u], alpha=0.8, stype='maxpcp', newfig=False)
    elif plot == 'displ':
        m.PlotContourDispl(U, n=np.c_[v, u], alpha=0.8, stype='mag', newfig=False)
    else:
        print('Unknown plot type in PlotMeshImage')
    # plt.xlim([0,f.pix.shape[1]])
    # plt.ylim([f.pix.shape[0],0])
    # plt.axis("on")

def PlotMeshImage3d(f, m, cam=None, U=None):
    """Plotting the mesh over the image. 

    Parameters
    ----------
    f : pyxel.Image
        The image over which to plot the mesh
    m : pyxel.Mesh
        The mesh
    cam : pyxel.Camera
        The camera model
    U : Numpy array
        A displacement dof vector (OPTIONNAL) to warp the mesh.
    plot : String (OPTIONNAL)
        'mesh': plot the mesh in yellow (DEFAULT)
        'displ': plot contour displacement field
        'strain': plot contour strain field

    """
    n = m.n.copy()
    if U is not None:
        n += U[m.conn]
    plt.figure()
    f.Plot()
    if cam is None:
        u, v, w = n[:, 0], n[:, 1], n[:, 2]
    else:
        u, v, w = cam.P(n[:, 0], n[:, 1], n[:, 2])
    # if volume elements, build surface mesh
    if list(m.e.keys())[0] in [5, 4, 11, 17]:
        mb = m.BuildBoundaryMesh()
    else:
        mb = m.Copy()
    mb.dim = 2
    plt.subplot(221)
    mb.Plot(n = np.c_[w, v], edgecolor='y', alpha=0.6)
    plt.subplot(223)
    mb.Plot(n = np.c_[w, u], edgecolor='y', alpha=0.6)
    plt.subplot(224)
    mb.Plot(n = np.c_[v, u], edgecolor='y', alpha=0.6)


def full_screen():
    # to set matplotlib figures in full screen (works for windows at least)
    figManager = plt.get_current_fig_manager()
    if hasattr(figManager, 'window'):
        if hasattr(figManager.window, 'showMaximized'):
            figManager.window.showMaximized()
        elif hasattr(figManager.window, 'maximize'):
            figManager.resize(figManager.window.maximize())
    plt.tight_layout()




import numpy as np


def Hooke(p, typc='isotropic_2D_ps'):
    """Compute Hooke tensor from elastic constants

    Parameters
    ----------
    p : Numpy Array
        p = [E, nu] for isotropic material
        p = [E1, E2, nu12, G12] for orthotropic material
    typc : string
        'isotropic_2D_ps' plane stress (DEFAULT)
        'isotropic_2D_pe' plane strain 
        'isotropic_2D_axi' axisymmetric 
        'orthotropic_2D'  
        'laminate_2D'
        'isotropic_3D'  
    el_set (OPTIONAL) required for heterogeneous samples
    it is a dict with keys as element type (like mesh.e) and value are id
    of the material type which refers to the dict of materials p
    
    Returns
    -------
    Numpy array
        Hooke tensor.

    """
    if typc == 'isotropic_2D_ps':
        E = p[0]
        v = p[1]
        return E / (1 - v**2) * np.array([[1, v, 0], [v, 1, 0], [0, 0, (1 - v) / 2]])
        # return E / (1 - v**2) * np.array([1, 1, 0.5*(1 - v), v, v])
    elif typc == 'isotropic_2D_pe':
        E = p[0]
        v = p[1]
        return E / ((1 + v)*(1 - 2*v)) * np.array([[1-v, v, 0], [v, 1-v, 0], [0, 0, (1 - 2*v) / 2]])
        # return E / ((1 + v)*(1 - 2*v)) * np.array([1-v, 1-v, 0.5*(1-2*v), v, v])
    elif typc == 'isotropic_2D_axi':
        E = p[0]
        v = p[1]
        return E / ((1 + v)*(1 - 2*v)) * np.array([[1-v, v, v, 0],
                                                    [v, 1-v, v, 0],
                                                    [v, v, 1-v, 0],
                                                    [0, 0, 0, (1 - 2*v) / 2]])
        # return E/((1+v)*(1-2*v)) * np.array([1-v, 1-v, 1-v, 0.5*(1-2*v), v, v, v, v, v, v])
    elif typc == 'orthotropic_2D':
        El = p[0]
        Et = p[1]
        vtl = p[2]
        Glt = p[3]
        vlt = vtl * El / Et
        alp = 1 / (1 - vlt * vtl)
        return np.array([[alp * El, alp * vtl * El, 0],
                        [alp * vlt * Et, alp * Et, 0],
                        [0, 0, 2 * Glt]])
    elif typc == 'laminate_2D':
        # mat.type='Laminate';
        # mat.El=165e9;
        # mat.Et=7.69e9;
        # mat.vlt=0.33;
        # mat.vtl=mat.vlt*mat.Et/mat.El;
        # mat.Glt=4.75e9;
        # mat.layup=[0 45 -45 -45 45 0
        #               ones(1,6)*0.1446*1e-3];

        # ep=mat.layup(2,:);
        # alpha=mat.layup(1,:);
        # El=mat.El;
        # Et=mat.Et;
        # vlt=mat.vlt;
        # vtl=mat.vtl;
        # Glt=mat.Glt;

        # % Orthotropic stiffness term in the ply coordinate syst
        # alp=1/(1-vlt*vtl);
        # H=[alp*El     alp*vtl*El 0
        #    alp*vlt*Et alp*Et     0
        #    0          0         2*Glt];

        # % rotation and addition of the ply properties
        # A=0*H;
        # for iply=1:numel(alpha)
        #     a=alpha(iply)*pi/180;
        #     c=cos(a);
        #     s=sin(a);
        #     T=[c^2 s^2  2*c*s
        #        s^2 c^2 -2*c*s
        #       -c*s c*s  (c^2-s^2)];
        #     Q=T*H*T';
        #     A=A+Q*ep(iply);
        # end
        # h=sum(ep);
        # hooke=A./h;
        raise Exception('HOOKE LAMINATE TODO')
    if typc == 'isotropic_3D':
        E = p[0]
        v = p[1]
        C = E / (2*(1+v)*(1-2*v)) * np.array([[2*(1-v), 2*v, 2*v, 0, 0, 0],
                                            [2*v, 2*(1-v), 2*v, 0, 0, 0],
                                            [2*v, 2*v, 2*(1-v), 0, 0, 0],
                                            [0, 0, 0, (1 - 2*v), 0, 0],
                                            [0, 0, 0, 0, (1 - 2*v), 0],
                                            [0, 0, 0, 0, 0, (1 - 2*v)]])
        return C
    else:
        raise Exception('Unknown elastic constitutive regime (3D)')


def Strain2Stress(hooke, En, Es):
    if len(hooke) == 3:  # dim 2
        SXX = En[:, 0] * hooke[0, 0] + En[:, 1] * \
            hooke[0, 1] + 2 * Es[:, 0] * hooke[0, 2]
        SYY = En[:, 0] * hooke[1, 0] + En[:, 1] * \
            hooke[1, 1] + 2 * Es[:, 0] * hooke[1, 2]
        SXY = En[:, 0] * hooke[2, 0] + En[:, 0] * \
            hooke[2, 1] + 2 * Es[:, 0] * hooke[2, 2]
        Sn = np.c_[SXX, SYY]
        Ss = np.c_[SXY, 0*SXY]
    else:  # dim 3
        i = 0
        SXX = En[:, 0] * hooke[i, 0] + En[:, 1] * hooke[i, 1] + \
            En[:, 2] * hooke[i, 2] + 2 * Es[:, 0] * hooke[i, 3] + \
            2 * Es[:, 1] * hooke[i, 4] + 2 * Es[:, 2] * hooke[i, 5]
        i = 1
        SYY = En[:, 0] * hooke[i, 0] + En[:, 1] * hooke[i, 1] +\
            En[:, 2] * hooke[i, 2] + 2 * Es[:, 0] * hooke[i, 3] +\
            2 * Es[:, 1] * hooke[i, 4] + 2 * Es[:, 2] * hooke[i, 5]
        i = 2
        SZZ = En[:, 0] * hooke[i, 0] + En[:, 1] * hooke[i, 1] +\
            En[:, 2] * hooke[i, 2] + 2 * Es[:, 0] * hooke[i, 3] +\
            2 * Es[:, 1] * hooke[i, 4] + 2 * Es[:, 2] * hooke[i, 5]
        i = 3
        SXY = En[:, 0] * hooke[i, 0] + En[:, 1] * hooke[i, 1] +\
            En[:, 2] * hooke[i, 2] + 2 * Es[:, 0] * hooke[i, 3] +\
            2 * Es[:, 1] * hooke[i, 4] + 2 * Es[:, 2] * hooke[i, 5]
        i = 4
        SXZ = En[:, 0] * hooke[i, 0] + En[:, 1] * hooke[i, 1] +\
            En[:, 2] * hooke[i, 2] + 2 * Es[:, 0] * hooke[i, 3] +\
            2 * Es[:, 1] * hooke[i, 4] + 2 * Es[:, 2] * hooke[i, 5]
        i = 5
        SYZ = En[:, 0] * hooke[i, 0] + En[:, 1] * hooke[i, 1] +\
            En[:, 2] * hooke[i, 2] + 2 * Es[:, 0] * hooke[i, 3] +\
            2 * Es[:, 1] * hooke[i, 4] + 2 * Es[:, 2] * hooke[i, 5]
        Sn = np.c_[SXX, SYY, SZZ]
        Ss = np.c_[SXY, SXZ, SYZ]
    return Sn, Ss


# %%
# import sympy as sp
# E, v = sp.symbols('E, v')
# Cinv = sp.Matrix([[1/E, -v/E, -v/E, 0, 0, 0],
#             [-v/E, 1/E, -v/E, 0, 0, 0],
#             [-v/E, -v/E, 1/E, 0, 0, 0],
#             [0, 0, 0, (1+v)/E, 0, 0],
#             [0, 0, 0, 0, (1+v)/E, 0],
#             [0, 0, 0, 0, 0, (1+v)/E]])
# C = Cinv**-1
# sp.pycode(C)




# -*- coding: utf-8 -*-
"""
Finite Element Digital Image Correlation method 

@author: JC Passieux, INSA Toulouse, 2021

pyxel

PYthon library for eXperimental mechanics using Finite ELements

"""
import os
import numpy as np
import scipy.sparse.linalg as splalg
import scipy.sparse as sps 
from .image import Image
from .camera import Camera, CameraVol
from .mesher import StructuredMesh, TetraMeshBox
#import matplotlib.pyplot as plt

#import pdb
#pdb.set_trace()
    
#%%
class DICEngine:
    def __init__(self):
        self.f = []
        self.wphiJdf = []
        self.dyn = []
        self.mean0 = []
        self.std0 = []

    def ComputeLHS(self, f, m, cam):
        """Compute the FE-DIC Left hand side operator with the modified GN
    
        Parameters
        ----------
        f : pyxel.Image
            Reference State Image
        m : pyxel.Mesh
            The FE mesh
        cam : pyxel.Camera
            Calibrated Camera model.
            
        Returns
        -------
        scipy sparse
            The DIC Hessian (in the modified GN sense)
    
        """
        if hasattr(f, "tck") == 0:
            f.BuildInterp()
        pgu, pgv = cam.P(m.pgx, m.pgy)
        self.f = f.Interp(pgu, pgv)
        fdxr, fdyr = f.InterpGrad(pgu, pgv)
        Jxx, Jxy, Jyx, Jyy = cam.dPdX(m.pgx, m.pgy)
        phiJdf = (
            sps.diags(fdxr * Jxx + fdyr * Jyx) @ m.phix
            + sps.diags(fdxr * Jxy + fdyr * Jyy) @ m.phiy
        )
        self.wphiJdf = sps.diags(m.wdetJ) @ phiJdf
        self.dyn = np.max(self.f) - np.min(self.f)
        self.mean0 = np.mean(self.f)
        self.std0 = np.std(self.f)
        self.f -= self.mean0
        return phiJdf.T @ self.wphiJdf

    def ComputeRHS(self, g, m, cam, U=[]):
        """Compute the FE-DIC right hand side operator with the modified GN
    
        Parameters
        ----------
        g : pyxel.Image
            Deformed State Image
        m : pyxel.Mesh
            The FE mesh
        cam : pyxel.Camera
            Calibrated Camera model.
        U : Numpy Array (OPTIONAL)
            The running displacement dof vector.
            
        Returns
        -------
        Numpy array
            DIC right hand side vector
        Numpy array
            The residual vector.
    
        """
        if hasattr(g, "tck") == 0:
            g.BuildInterp()
        if len(U) != m.ndof:
            U = np.zeros(m.ndof)
        u, v = cam.P(m.pgx + m.phix @ U, m.pgy + m.phiy @ U)
        res = g.Interp(u, v)
        res -= np.mean(res)
        std1 = np.std(res)
        res = self.f - self.std0 / std1 * res
        B = self.wphiJdf.T @ res
        return B, res

    def ComputeLHS_EB(self, f, m, cam):
        """Compute the FE-DIC Left hand side operator with the modified GN
        and with elementary correction of brigthness and contrast
    
        Parameters
        ----------
        f : pyxel.Image
            Reference State Image
        m : pyxel.Mesh
            The FE mesh
        cam : pyxel.Camera
            Calibrated Camera model.
            
        Returns
        -------
        scipy sparse
            The DIC Hessian (in the modified GN sense)
    
        """
        if hasattr(f, "tck") == 0:
            f.BuildInterp()
        pgu, pgv = cam.P(m.pgx, m.pgy)
        self.f = f.Interp(pgu, pgv)
        fdxr, fdyr = f.InterpGrad(pgu, pgv)
        Jxx, Jxy, Jyx, Jyy = cam.dPdX(m.pgx, m.pgy)
        phiJdf = (
            sps.diags(fdxr * Jxx + fdyr * Jyx) @ m.phix
            + sps.diags(fdxr * Jxy + fdyr * Jyy) @ m.phiy
        )
        self.wphiJdf = sps.diags(m.wdetJ) @ phiJdf
        self.dyn = np.max(self.f) - np.min(self.f)
        ff = sps.diags(self.f) @ m.Me
        mean0 = np.asarray(np.mean(ff, axis=0))[0]
        self.std0 = np.asarray(np.sqrt(np.mean(ff.power(2), axis=0) - mean0 ** 2))[0]
        self.f -= m.Me @ mean0.T
        return phiJdf.T @ self.wphiJdf

    def ComputeRHS_EB(self, g, m, cam, U=[]):
        """Compute the FE-DIC right hand side operator with the modified GN
        and with elementary correction of brigthness and contrast
    
        Parameters
        ----------
        g : pyxel.Image
            Deformed State Image
        m : pyxel.Mesh
            The FE mesh
        cam : pyxel.Camera
            Calibrated Camera model.
        U : Numpy Array (OPTIONAL)
            The running displacement dof vector.
            
        Returns
        -------
        Numpy array
            DIC right hand side vector
        Numpy array
            The residual vector.
        """
        if hasattr(g, "tck") == 0:
            g.BuildInterp()
        if len(U) != m.ndof:
            U = np.zeros(m.ndof)
        pgxu = m.pgx + m.phix.dot(U)
        pgyv = m.pgy + m.phiy.dot(U)
        u, v = cam.P(pgxu, pgyv)
        res = g.Interp(u, v)
        ff = sps.diags(res).dot(m.Me)
        mean0 = np.asarray(np.mean(ff, axis=0))[0]
        std0 = np.asarray(np.sqrt(np.mean(ff.power(2), axis=0) - mean0 ** 2))[0]
        res -= m.Me @ mean0
        res = self.f - sps.diags(m.Me @ (self.std0 / std0)) @ res
        B = self.wphiJdf.T @ res
        return B, res
            
    def ComputeLHS2(self, f, g, m, cam, U):
        """Compute the FE-DIC right hand side operator with the true Gauss Newton
    
        Parameters
        ----------
        f : pyxel.Image
            Reference State Image
        g : pyxel.Image
            Deformed State Image
        m : pyxel.Mesh
            The FE mesh
        cam : pyxel.Camera
            Calibrated Camera model.
        U : Numpy Array (OPTIONAL)
            The running displacement dof vector.
            
        Returns
        -------
        scipy sparse
            The DIC Hessian (in the GN sense)
    
        """
        if hasattr(f, "tck") == 0:
            f.BuildInterp()
        if hasattr(g, "tck") == 0:
            g.BuildInterp()
        pgu, pgv = cam.P(m.pgx, m.pgy)
        self.f = f.Interp(pgu, pgv)
        pgu, pgv = cam.P(m.pgx + m.phix.dot(U), m.pgy + m.phiy.dot(U))
        fdxr, fdyr = g.InterpGrad(pgu, pgv)
        Jxx, Jxy, Jyx, Jyy = cam.dPdX(m.pgx, m.pgy)
        phiJdf = sps.diags(fdxr * Jxx + fdyr * Jyx).dot(m.phix) + sps.diags(
            fdxr * Jxy + fdyr * Jyy
        ).dot(m.phiy)
        self.wphiJdf = sps.diags(m.wdetJ).dot(phiJdf)
        self.dyn = np.max(self.f) - np.min(self.f)
        self.mean0 = np.mean(self.f)
        self.std0 = np.std(self.f)
        self.f -= self.mean0
        return phiJdf.T.dot(self.wphiJdf)

    def ComputeRHS2(self, g, m, cam, U=[]):
        """Compute the FE-DIC right hand side operator with the true Gauss-Newton
    
        Parameters
        ----------
        g : pyxel.Image
            Deformed State Image
        m : pyxel.Mesh
            The FE mesh
        cam : pyxel.Camera
            Calibrated Camera model.
        U : Numpy Array (OPTIONAL)
            The running displacement dof vector.
            
        Returns
        -------
        Numpy array
            DIC right hand side vector
        Numpy array
            The residual vector.
    
        """
        if hasattr(g, "tck") == 0:
            g.BuildInterp()
        if len(U) != m.ndof:
            U = np.zeros(m.ndof)
        u, v = cam.P(m.pgx + m.phix @ U, m.pgy + m.phiy @ U)
        res = g.Interp(u, v)
        res -= np.mean(res)
        std1 = np.std(res)
        res = self.f - self.std0 / std1 * res
        fdxr, fdyr = g.InterpGrad(u, v)
        Jxx, Jxy, Jyx, Jyy = cam.dPdX(m.pgx, m.pgy)
        wphiJdf = (
            sps.diags(m.wdetJ * (fdxr * Jxx + fdyr * Jyx)) @ m.phix
            + sps.diags(m.wdetJ * (fdxr * Jxy + fdyr * Jyy)) @ m.phiy
        )
        B = wphiJdf.T @ res
        return B, res

    def ComputeRES(self, g, m, cam, U=None):
        """Compute the FE-DIC residual
    
        Parameters
        ----------
        g : pyxel.Image
            Deformed State Image
        m : pyxel.Mesh
            The FE mesh
        cam : pyxel.Camera
            Calibrated Camera model.
        U : Numpy Array (OPTIONAL)
            The displacement dof vector.

        Returns
        -------
        Numpy array
            the residual vector.
    
        """
        if hasattr(g, "tck") == 0:
            g.BuildInterp()
        if U is None:
            U = np.zeros(m.ndof)
        pgxu = m.pgx + m.phix.dot(U)
        pgyv = m.pgy + m.phiy.dot(U)
        u, v = cam.P(pgxu, pgyv)
        res = g.Interp(u, v)
        res -= np.mean(res)
        std1 = np.std(res)
        res = self.f - self.std0 / std1 * res
        return res

#%% 

class DVCEngine: 
    def __init__(self):        
        self.f = []
        self.wphiJdf = []
        self.dyn = []
        self.mean0 = []
        self.std0 = []
        
    def ComputeLHS(self, f, m, cam):
        """Compute the FE-DIC Left hand side operator with the modified GN
    
        Parameters
        ----------
        f : Image object 
            Reference State Image
        m : Mesh object 
            The FE mesh
            
        Returns
        -------
        scipy sparse
            The DIC Hessian (in the modified GN sense)
    
        """
        pgu, pgv, pgw = cam.P(m.pgx, m.pgy, m.pgz)
        self.f = f.Interp(pgu, pgv, pgw)
        fdxr, fdyr, fdzr = f.InterpGrad(pgu, pgv, pgw)
        Jxx, Jxy, Jxz, Jyx, Jyy, Jyz, Jzx, Jzy, Jzz = cam.dPdX(m.pgx, m.pgy, m.pgz)
        phiJdf = (
            sps.diags(fdxr * Jxx + fdyr * Jyx + fdzr * Jzx) @ m.phix + \
            sps.diags(fdxr * Jxy + fdyr * Jyy + fdzr * Jzy) @ m.phiy + \
            sps.diags(fdxr * Jxz + fdyr * Jyz + fdzr * Jzz) @ m.phiz 
        )
        self.wphiJdf = sps.diags(m.wdetJ) @ phiJdf
        self.dyn = np.max(self.f) - np.min(self.f)
        self.mean0 = np.mean(self.f)
        self.std0  = np.std(self.f)
        self.f     -= self.mean0
        return phiJdf.T @ self.wphiJdf
        
        # self.f = f.Interp( m.pgx , m.pgy, m.pgz)
        # dfdx, dfdy, dfdz = f.InterpGrad( m.pgx, m.pgy, m.pgz )
        # phidf = sps.diags(dfdx).dot(m.phix ) + sps.diags(dfdy).dot(m.phiy) + sps.diags(dfdz).dot(m.phiz)
        # self.wphidf = sps.diags(m.wdetJ) @ phidf 
        # self.dyn = np.max(self.f) - np.min(self.f)
        # self.mean0 = np.mean(self.f)
        # self.std0  = np.std(self.f)
        # self.f     -= self.mean0
        # return phidf.T @ self.wphidf

    def ComputeRHS(self, g, m, cam=None, U=None): 
        """Compute the FE-DIC right hand side operator with the modified GN
    
        Parameters
        ----------
        g : pyxel.Image
            Deformed State Image
        m : pyxel.Mesh
            The FE mesh
        U : Numpy Array (OPTIONAL)
            The running displacement dof vector.
            
        Returns
        -------
        Numpy array
            DIC right hand side vector
        Numpy array
            The residual vector.
    
        """
        if U is None:
            U = np.zeros(m.ndof)
        x =  m.pgx + m.phix @ U    
        y =  m.pgy + m.phiy @ U 
        z =  m.pgz + m.phiz @ U 
        #
        pgu, pgv, pgw = cam.P(x, y, z)
        res = g.Interp(pgu, pgv, pgw)
        # res = g.Interp(x,y,z) 
        res -= np.mean(res)
        std1 = np.std(res)
        res = self.f - self.std0 / std1 * res
        # b = self.wphidf.T @ res
        b = self.wphiJdf.T @ res
        return b, res
    
    def ComputeRES(self, g, m, cam=None, U=None):
        """Compute the FE-DIC residual
    
        Parameters
        ----------
        g : pyxel.Image
            Deformed State Image
        m : pyxel.Mesh
            The FE mesh
        U : Numpy Array (OPTIONAL)
            The displacement dof vector.
        Returns
        -------
        Numpy array
            the residual vector.
    
        """
        if U is None:
            U = np.zeros(m.ndof)
        x = m.pgx + m.phix.dot(U)
        y = m.pgy + m.phiy.dot(U)
        z = m.pgz + m.phiz.dot(U)  
        pgu, pgv, pgw = cam.P(x, y, z)
        res = g.Interp(pgu, pgv, pgw)
        res -= np.mean(res)
        std1 = np.std(res)
        res = self.f - self.std0 / std1 * res
        return res            

#%% 

def MeshFromROI(roi, dx, typel=3):
    """Build a structured FE mesh and a pyxel.camera object from a region
    of interest (ROI) selected in an image f

    Parameters
    ----------
    roi : numpy.array
        The Region of Interest made using  f.SelectROI(), f being a pyxel.Image
    dx : numpy or python array
        dx  = [dx, dy]: average element size (can be scalar) in pixels
    typel : int
        type of element: {3: 'qua4',2: 'tri3',9: 'tri6',16: 'qua8',10: 'qua9'}

    Returns
    -------
    m : pyxel.Mesh
        The finite element mesh
    cam : pyxel.Camera
        The corresponding camera

    Example:
    -------
    f.SelectROI()  -> select the region with rectangle selector
                      and copy - paste the roi in the python terminal
    m, cam = px.MeshFromROI(roi, [20, 20], 3)
    """
    if typel in [4, 5]: #volume elements
        if typel == 4:
            m = TetraMeshBox(roi, dx)
        else:
            m = StructuredMesh(roi, dx, typel=typel)
        # xm = np.mean(m.n, axis=0)
        # m.n -= xm[np.newaxis]
        # cam = CameraVol([1, xm[0], xm[1], xm[2], 0, 0, 0])
        cam = CameraVol([1, 0, 0, 0, 0, 0, 0])
        return m, cam
    else:
        m = StructuredMesh(roi, dx, typel=typel)
        m.n[:,1] *= -1
        p = np.array([1., 0., 0., 0.])
        cam = Camera(p)
        return m, cam

def Correlate(f, g, m, cam, dic=None, H=None, U0=None, l0=None, Basis=None, 
              L=None, eps=None, maxiter=30, disp=True, EB=False, direct=False):
    """Perform FE-Digital Image Correlation.

    Parameters
    ----------
    f : pyxel.Image
        Reference Image
    g : pyxel.Image
        Deformed State Image
    m : pyxel.Mesh
        The FE mesh
    cam : pyxel.Camera
        Calibrated Camera model.
    dic : pyxel.DICEngine (OPTIONAL)
        An existing DIC engine where ComputeLHS is pre-computed.
        Allow to perform multi time step correlation faster. 
    H : scipy sparse (OPTIONAL)
        DIC Hessian operator (avoid recomputing when constant)
    U0 : Numpy Array (OPTIONAL)
        Initial guess for the displacement dof vector.
    l0 : float (OPTIONAL)
        regularization length in physical (mesh) unit
    Basis : Numpy array (OPTIONAL)
        Reduced basis for use in iDIC for instance
    L : scipy sparse (OPTIONAL)
        Regularization operator, for instance computed with L = pyxel.Laplacian()
    eps : float (OPTIONAL)
        stopping criterion for dU/U
    disp : Bool (DEFAULT=True)
        Display error and residual magnitude at each iteration.
    EB: Element-wise brightness and contrast correction. Default: False means
        homogeneous B&C correction
        
    Returns
    -------
    Numpy array
        Displacement DOF vector
    Numpy array
        Residual vector

    """
    analysis = 'dic'
    if len(f.pix.shape) == 3:
        analysis = 'dvc'
    if dic is None:
        if analysis == 'dvc':
            dic = DVCEngine()
        else:
            dic = DICEngine()
    if len(m.conn) == 0:
        m.Connectivity()
    if U0 is None:
        U = np.zeros(m.ndof)
    else:
        U = U0.copy()
    if m.phix is None:
        if analysis == 'dvc':
            m.DVCIntegration(cam)
        else:
            m.DICIntegration(cam, EB=EB)
    if H is None:
        if EB:
            H = dic.ComputeLHS_EB(f, m, cam)
        else:
            H = dic.ComputeLHS(f, m, cam)
    if l0 == 0:
        l0 = None
    if eps is None:
        eps = 1e-3
    no_reg = True
    if l0 is not None:
        # Weak regularisation
        no_reg = False
        if L is None:
            L = m.Laplacian()
        T = 10 * m.GetApproxElementSize()
        V = m.PlaneWave(T)
        H0 = V.dot(H.dot(V))
        L0 = V.dot(L.dot(V))
        ll = (l0/T)**2 * H0 / L0
        print('Regularization: Weak with λ = %2.3f' % ll)
        Hfull = H + ll * L
    else:
        Hfull = H.copy()
    if Basis is not None:
        # Strong Regularization
        no_reg = False
        print('Regularization: Strong with reduced basis')
        Hfull = Basis.T @ Hfull @ Basis
    if no_reg:
        print('Regularization: None')
    if direct:
        print('Factorization...')
        Hfull = splalg.splu(Hfull)
    else:
        # DIAGONAL PRECONDITIONER
        print('Jacobi Preconditioner...')
        eps_zero = 1e-5 * np.min(Hfull)
        Mfull = sps.diags(1/(Hfull.diagonal() + eps_zero))
        # ILU PRECONDITIONER
        # print('Incomplete LU Factorization...')
        # MLU = splalg.spilu(Hfull, drop_tol=0.1)
        # M = lambda x: MLU.solve(x)
        # Mfull = splalg.LinearOperator(Hfull.shape, M)
    stdr_old = 100
    for ik in range(0, maxiter):
        if EB:
            [b, res] = dic.ComputeRHS_EB(g, m, cam, U)
        else:
            [b, res] = dic.ComputeRHS(g, m, cam, U)
        if l0 is not None:
            b -= ll * L @ U
        if Basis is not None:
            b = Basis.T @ b
        if direct:
            dU = Hfull.solve(b)
        else:
            dU, info = splalg.cg(Hfull, b, tol=0.1, M=Mfull)
            if info:
                print('Iterative solver did not reach convergence!')
        if Basis is not None:
            dU = Basis @ dU
        U += dU
        err = np.linalg.norm(dU) / np.linalg.norm(U)
        stdr = np.std(res)
        if disp:
            print("Iter # %2d | std(res)=%2.2f gl | dU/U=%1.2e" % (ik + 1, stdr, err))
        if err < eps:
            # if err < eps or abs(stdr - stdr_old) < 1e-3:
            break
        stdr_old = stdr
    return U, res

def MultiscaleInit(imf, img, m, cam, scales=[3, 2, 1], l0=None, U0=None,
                   Basis=None, eps=None, disp=True, direct=True):
    """Perform Multigrid initialization for FE-Digital Image Correlation.

    Parameters
    ----------
    f : pyxel.Image
        Reference Image
    g : pyxel.Image
        Deformed State Image
    m : pyxel.Mesh
        The FE mesh
    cam : pyxel.Camera
        Calibrated Camera model.
    scales : python list (DEFAULT=[3,2,1])
        An ordered list of scales for the multigrid initialization.
        Each time image is subsampled by 2**scale.
        Scale 0 correspond to initial image
    l0 : float (OPTIONAL)
        regularization length in physical (mesh) unit
        - set l0 to None to automatically compute l0
        - set l0 to 0 to descativate regularization in the multiscale process
    U0 : Numpy Array (OPTIONAL)
        Initial guess for the displacement dof vector.
    Basis : Numpy array (OPTIONAL)
        Reduced basis for use in iDIC for instance
    L : scipy sparse (OPTIONAL)
        Regularization operator, for instance computed with L = pyxel.Laplacian()
    eps : float (OPTIONAL)
        stopping criterion for dU/U
    disp : Bool (DEFAULT=True)
        Display error and residual magnitude at each iteration.
        
    Returns
    -------
    Numpy array
        Displacement DOF vector

    """
    if len(m.conn) == 0:
        m.Connectivity()
    # estimate average element size in pixels
    aes = int(m.GetApproxElementSize(cam))
    print('Average Element Size in px: %3d' % aes)
    if l0 is None:
        if Basis is None:
            if cam is None:
                l0 = 30
            else:
                l0 = 30/cam.get_p()[0]
            print('Auto reg. length l0 = %2.3e' % l0)
        else:
            l0 = 0
        L = None
    else:
        L = m.Laplacian()
    if U0 is None:
        U = np.zeros(m.ndof)
    else:
        U = U0.copy()
    for js in range(len(scales)):
        iscale = scales[js]
        if disp:
            print("SCALE %2d" % (iscale))
        f = imf.Copy()
        f.SubSample(iscale)
        g = img.Copy()
        g.SubSample(iscale)
        if cam is not None:
            cam2 = cam.SubSampleCopy(iscale)
        else:
            cam2 = None
        m2 = m.Copy()
        if len(f.pix.shape) == 3:
            aesi = min(5, aes // (2**iscale))  # max 5 integration points > Fast
            aesi = max(1, aesi)  # not smaller than 1
            m2.DVCIntegration(aesi)
        else:
            aes2 = max(aes // (2**iscale), 2)
            m2.DICIntegrationFast(aes2)
        # m2.DICIntegration(cam2)
        # import matplotlib.pyplot as plt
        # plt.figure()
        # g.Plot()
        # u, v = cam2.P(m2.n[:, 0], m2.n[:, 1])
        # m.Plot(n=np.c_[v, u], edgecolor="y", alpha=0.6)
        # plt.figure()
        # m2.Plot()
        # plt.plot(m2.pgx,m2.pgy,'k.')
        # plt.axis('equal')
        # from .utils import PlotMeshImage3d
        # PlotMeshImage3d(f, m2, cam2)
        U, r = Correlate(f, g, m2, cam2, l0=l0*2**iscale, Basis=Basis,
                         L=L, U0=U, eps=eps, disp=disp, direct=direct)
   
        # PlotMeshImage3d(g, m2, cam2, U)
    return U


def CorrelateTimeIncr(m, f, imagefile, imnums, cam, scales):
    """Performs FE-DIC for a time image series.

    Parameters
    ----------
    m : pyxel.Mesh
        The FE mesh
    f : pyxel.Image
        Reference Image
    imagefile : string
        a generic filename for the deformed state images.
        example: imagefile = os.path.join('data', 'dic_composite', 'zoom-0%03d_1.tif')
        such that imagefile % 30 is the filename 'data/dic_composite/zoom-0030_1.tif'
    imnums : Numpy Array
        The array containing the deformed state image numbers
    cam : pyxel.Camera
        Calibrated Camera model.
    scales : python list (DEFAULT=[3,2,1])
        An ordered list of scales for the multigrid initialization.
        Each time image is subsampled by 2**scale.
        Scale 0 correspond to initial image
        
    Returns
    -------
    Numpy array
        An Array containing the displacement DOF vector, one column for one timestep.

    """
    UU = np.zeros((m.ndof, len(imnums)))
    if len(m.pgx) == 0:
        m.DICIntegration(cam)
    dic = DICEngine()
    H = dic.ComputeLHS(f, m, cam)
    im = 1
    print(" ==== IMAGE %3d === " % imnums[im])
    imdef = imagefile % imnums[im]
    g = Image(imdef).Load()
    UU[:, im] = MultiscaleInit(f, g, m, cam, scales=scales)
    UU[:, im], r = Correlate(f, g, m, cam, dic=dic, H=H, U0=UU[:, im])
    for im in range(2, len(imnums)):
        print(" ==== IMAGE %3d === " % imnums[im])
        imdef = imagefile % imnums[im]
        g = Image(imdef).Load()
        if True:
            UU[:, im] = MultiscaleInit(
                f, g, m, cam, scales=scales, U0=UU[:, im - 1], eps=1e-4
            )
            UU[:, im], r = Correlate(f, g, m, cam, dic=dic, H=H, U0=UU[:, im], eps=1e-4)
        else:
            V = UU[:, [im - 1]]
            UU[:, im] = MultiscaleInit(
                f, g, m, cam, scales=scales, Basis=V, U0=UU[:, im - 1], eps=1e-4
            )
            UU[:, im], r = Correlate(
                f, g, m, cam, dic=dic, H=H, Basis=V, U0=UU[:, im], eps=1e-4
            )
            UU[:, im], r = Correlate(f, g, m, cam, dic=dic, H=H, U0=UU[:, im], eps=1e-4)
        if not os.path.isdir('tmp'):
            os.makedirs('tmp')
        np.save(os.path.join('tmp', 'multiscale_init_tmp'), UU)
    return UU


def DISFlowInit(imf, img, m=None, cam=None, meth='MEDIUM'):
    """
    Compute initial guess using OpenCV DISFlow routine

    Parameters
    ----------
    imf : PYXEL.IMAGE
        Reference image
    img : PYXEL.IMAGE
        Deformed state image
    m : PYXEL.MESH
        finite element mesh
        if None > return the result of DISFlow
    cam : PYXEL.CAMERA
        Camera model
        if None > return the result of DISFlow
    meth : STRING, optional
        'MEDIUM': medium option of DISFlow
        'FAST': fast option of DISFlow
        'ULTRAFAST': ultrafast option of DISFlow
        otherwise: manual settings for DISFlow
        DESCRIPTION. The default is 'MEDIUM'.

    Returns
    -------
    u : NUMPY.ARRAY
        initial guess DOF vector if m and cam are given
        returns pixmaps U, V is cam is None

    """
    import cv2
    if meth == 'MEDIUM':
        flow=cv2.DISOpticalFlow_create(cv2.DISOPTICAL_FLOW_PRESET_MEDIUM)
    elif meth == 'FAST':
        flow=cv2.DISOpticalFlow_create(cv2.DISOPTICAL_FLOW_PRESET_FAST)
    elif meth == 'ULTRAFAST':
        flow=cv2.DISOpticalFlow_create(cv2.DISOPTICAL_FLOW_PRESET_ULTRAFAST)
    else:
        # MANUAL
        flow = cv2.DISOpticalFlow_create()
        flow.setVariationalRefinementAlpha(20.0)		# Laplacian of displacment
        flow.setVariationalRefinementGamma(10.0)		# Gradient of image consistency
        flow.setVariationalRefinementDelta(5.0) 	    # Optical flow
        flow.setVariationalRefinementIterations(5)	    # Number of iterations
        flow.setFinestScale(0)
        flow.setPatchSize(13)
        flow.setPatchStride(1)
    UV = flow.calc(imf.pix.astype('uint8'), img.pix.astype('uint8'), None)
    U = UV[::,::,0]
    V = UV[::,::,1]
    if m is None:
        return U, V
    else: 
        u, v = cam.P(m.n[:,0],m.n[:,1])
        fp = imf.Copy()
        fp.pix = V
        fp.BuildInterp()
        du = fp.Interp(u, v)
        fp.pix = U
        fp.BuildInterp()
        dv = fp.Interp(u, v)
        
        Xdx, Ydy = cam.PinvNL(u+du, v+dv)
        Ux = Xdx - m.n[:,0]
        Uy = Ydy - m.n[:,1]
        
        if len(m.conn) == 0 :
            m.Connectivity()
        u = np.zeros(m.ndof)
        u[m.conn[:,0]] = Ux
        u[m.conn[:,1]] = Uy
        return u


# -*- coding: utf-8 -*-
"""
Finite Element Digital Image Correlation method

@author: A. Rouawne, INSA Toulouse, 2023

pyxel

PYthon library for eXperimental mechanics using Finite ELements

"""
import os
import numpy as np
import scipy as sp
from .bspline_routines import bspdegelev, bspkntins, global_basisfuns, Get2dBasisFunctionsAtPts, global_basisfunsWd
import scipy.sparse as sps
import matplotlib.pyplot as plt
from .camera import Camera
from .mesher import StructuredMeshQ4


class BSplinePatch(object):
    def __init__(self, ctrlPts, degree, knotVect):
        """
        Nurbs surface from R^2 (xi,eta)--->R^2 (x,y)
        ctrlPts = [X,Y]
        """
        self.dim = 2
        self.ctrlPts = ctrlPts
        self.n = self.CrtlPts2N()
        self.degree = np.array(degree)
        self.knotVect = knotVect

        self.ien = 0              # NURBS Connectivity p-uplet (IENu,IENv)
        # Connectivity: Control Point number (column) of each element (line)
        self.noelem = 0
        self.tripleien = 0        # Correspondance between 2D elements and 1D elements
        self.iperM = 0            # Sparse matrix for coincident control points

        # Dictionary containing the value det(Jacobian)*Gauss_Weights*ElementMeasure/4
        self.wdetJmes = 0
        # Dictionary containing the measure of the elements in the ordre of listeltot
        self.mes = 0

        # Dictionary of basis functions evaluated at gauss points
        self.phi = np.empty(0)
        # Dictionary containing the derivative of Basis function in x direction
        self.dphidx = np.empty(0)
        # Dictionary containing the derivative of Basis function in y direction
        self.dphidy = np.empty(0)

        """ Attributes when using vectorization  """
        """ In this case, the implicit connectivity of the structured B-spline parametric space is used """
        self.npg = 0
        self.phix = np.empty(0)      # Matrix (N,0)
        self.phiy = np.empty(0)      # Matrix (0,N)
        self.dphixdx = np.empty(0)  # Matrix (dNdx,0)
        self.dphixdy = np.empty(0)  # Matrix (dNdy,0)
        self.dphiydx = np.empty(0)  # Matrix (0,dNdx)
        self.dphiydy = np.empty(0)  # Matrix (0,dNdy)
        self.wdetJ = np.empty(0)        # Integration weights diagonal matrix
        self.pgx = np.empty(0)
        self.pgy = np.empty(0)

        self.phiMatrix = 0
        self.n_elems = 0
        self.pix = 0
        self.piy = 0
        self.piz = 0
        self.integrationCellsCoord = 0
        """ fixed parameters for integration """
        self.nbg_xi = 0
        self.nbg_eta = 0
        self.Gauss_xi = 0
        self.Gauss_eta = 0
        self.wgRef = 0
        self.refGaussTriangle = 0

    def Copy(self):
        m = BSplinePatch(self.ctrlPts.copy(), self.degree.copy(),
                         self.knotVect.copy())
        m.conn = self.conn.copy()
        m.ndof = self.ndof
        m.dim = self.dim
        m.npg = self.npg
        m.pgx = self.pgx.copy()
        m.pgy = self.pgy.copy()
        m.phix = self.phix.copy()
        m.phiy = self.phiy.copy()
        m.wdetJ = self.wdetJ.copy()
        return m

    def IsRational(self):
        return (self.ctrlPts[3] != 1).any()

    def Get_nbf_1d(self):
        """ Get the number of basis functions per parametric direction """
        return self.ctrlPts.shape[1:]

    def Get_nbf(self):
        """ Total number of basis functions """
        return np.product(self.Get_nbf_1d())

    def Get_nbf_elem_1d(self):
        return self.degree + 1

    def Get_nbf_elem(self):
        return np.product(self.degree+1)

    def Get_listeltot(self):
        """ Indices of elements """
        return np.arange(self.ien[0].shape[1]*self.ien[1].shape[1])

    def Get_P(self):
        """  Returns the total"""
        P = np.c_[self.ctrlPts[0].ravel(order='F'),
                  self.ctrlPts[1].ravel(order='F')]
        return P

    def SelectNodes(self, n=-1):
        """
        Selection of nodes by hand in a mesh.
        """
        plt.figure()
        self.Plot()
        figManager = plt.get_current_fig_manager()
        if hasattr(figManager.window, 'showMaximized'):
            figManager.window.showMaximized()
        else:
            if hasattr(figManager.window, 'maximize'):
                figManager.resize(figManager.window.maximize())
        plt.title("Select " + str(n) + " points... and press enter")
        pts1 = np.array(plt.ginput(n, timeout=0))
        plt.close()
        dx = np.kron(np.ones(pts1.shape[0]), self.n[:, [0]]) - np.kron(
            np.ones((self.n.shape[0], 1)), pts1[:, 0]
        )
        dy = np.kron(np.ones(pts1.shape[0]), self.n[:, [1]]) - np.kron(
            np.ones((self.n.shape[0], 1)), pts1[:, 1]
        )
        nset = np.argmin(np.sqrt(dx ** 2 + dy ** 2), axis=0)
        self.Plot()
        plt.plot(self.n[nset, 0], self.n[nset, 1], "ro")
        return nset

    def SelectLine(self, eps=1e-8):
        """
        Selection of the nodes along a line defined by 2 nodes.
        """
        plt.figure()
        self.Plot()
        figManager = plt.get_current_fig_manager()
        if hasattr(figManager.window, 'showMaximized'):
            figManager.window.showMaximized()
        else:
            if hasattr(figManager.window, 'maximize'):
                figManager.resize(figManager.window.maximize())
        plt.title("Select 2 points of a line... and press enter")
        pts1 = np.array(plt.ginput(2, timeout=0))
        plt.close()
        n1 = np.argmin(np.linalg.norm(self.n - pts1[0, :], axis=1))
        n2 = np.argmin(np.linalg.norm(self.n - pts1[1, :], axis=1))
        v = np.diff(self.n[[n1, n2]], axis=0)[0]
        nv = np.linalg.norm(v)
        v = v / nv
        n = np.array([v[1], -v[0]])
        c = n.dot(self.n[n1, :])
        (rep,) = np.where(abs(self.n.dot(n) - c) < eps)
        c1 = v.dot(self.n[n1, :])
        c2 = v.dot(self.n[n2, :])
        nrep = self.n[rep, :]
        (rep2,) = np.where(((nrep.dot(v) - c1)
                            * (nrep.dot(v) - c2)) < nv * 1e-2)
        nset = rep[rep2]
        self.Plot()
        plt.plot(self.n[nset, 0], self.n[nset, 1], "ro")
        return nset

    def Connectivity(self):
        nn = len(self.n)
        self.ndof = nn * self.dim
        self.conn = np.c_[np.arange(nn), np.arange(nn) + nn]

    def CrtlPts2N(self, ctrlPts=None):
        if ctrlPts is None:
            ctrlPts = self.ctrlPts.copy()
        n = np.c_[ctrlPts[0].ravel(order='F'),
                  ctrlPts[1].ravel(order='F')]
        return n

    def N2CrtlPts(self, n=None):
        # n should be in the right order (xi, eta) meshgrid
        if n is None:
            n = self.n.copy()
        nbf = self.Get_nbf_1d()
        ctrlPts = np.array([n[:, 0].reshape(nbf, order='F'),
                            n[:, 1].reshape(nbf, order='F')])
        return ctrlPts

    def BS2FE(self, U, n=[30, 30]):
        xi = np.linspace(self.knotVect[0][self.degree[0]],
                         self.knotVect[0][-self.degree[0]], n[0])
        eta = np.linspace(self.knotVect[1][self.degree[1]],
                          self.knotVect[1][-self.degree[1]], n[1])
        phi, _, _ = self.ShapeFunctionsAtGridPoints(xi, eta)
        x = phi.dot(self.n[:, 0])
        y = phi.dot(self.n[:, 1])
        roi = np.c_[np.ones(2), n].T-1
        mfe = StructuredMeshQ4(roi, 1)
        mfe.n = np.c_[x, y]
        # mfe.Plot()
        mfe.Connectivity()
        V = np.zeros(mfe.ndof)
        V[mfe.conn[:, 0]] = phi.dot(U[self.conn[:, 0]])
        V[mfe.conn[:, 1]] = phi.dot(U[self.conn[:, 1]])
        # mfe.Plot(V*30)
        return mfe, V, phi

    def VTKSol(self, filename, U=None, n=[30, 30]):
        # Surface
        if U is None:
            U = np.zeros(self.ndof)
        m, V, phi = self.BS2FE(U, n)
        m.VTKSol(filename, V)
        # Control mesh
        nbf = self.Get_nbf_1d()
        roi = np.c_[np.ones(2, dtype=int), nbf].T-1
        mfe = StructuredMeshQ4(roi, 1)
        mfe.n = np.c_[self.ctrlPts[0].ravel(),
                      self.ctrlPts[1].ravel()]
        mfe.Connectivity()
        V = U.copy()
        V[self.conn[:, 0]] = U[self.conn[:, 0]].reshape(nbf, order='F').ravel()
        V[self.conn[:, 1]] = U[self.conn[:, 1]].reshape(nbf, order='F').ravel()
        mfe.VTKSol(filename+'_cp', V)

    def Plot(self, U=None, n=None, neval=[30, 30], **kwargs):
        """ Physical elements = Image of the parametric elements on Python """
        alpha = kwargs.pop("alpha", 1)
        edgecolor = kwargs.pop("edgecolor", "k")
        nbf = self.Get_nbf()
        if n is None:
            n = self.Get_P()  # control points
        if U is None:
            U = np.zeros(2*nbf)
        Pxm = n[:, 0] + U[:nbf]
        Pym = n[:, 1] + U[nbf:]

        xi = np.linspace(
            self.knotVect[0][self.degree[0]], self.knotVect[0][-self.degree[0]], neval[0])
        eta = np.linspace(
            self.knotVect[1][self.degree[1]], self.knotVect[1][-self.degree[1]], neval[1])
        # Iso parameters for the elemnts
        xiu = np.unique(self.knotVect[0])
        etau = np.unique(self.knotVect[1])

        # Basis functions
        phi_xi1 = global_basisfunsWd(self.degree[0], self.knotVect[0], xiu)
        phi_eta1 = global_basisfunsWd(self.degree[1], self.knotVect[1], eta)
        phi_xi2 = global_basisfunsWd(self.degree[0], self.knotVect[0], xi)
        phi_eta2 = global_basisfunsWd(self.degree[1], self.knotVect[1], etau)

        phi1 = sps.kron(phi_eta1,  phi_xi1,  'csc')
        phi2 = sps.kron(phi_eta2,  phi_xi2,  'csc')

        xe1 = phi1.dot(Pxm)
        ye1 = phi1.dot(Pym)
        xe2 = phi2.dot(Pxm)
        ye2 = phi2.dot(Pym)

        xe1 = xe1.reshape((xiu.size, neval[1]), order='F')
        ye1 = ye1.reshape((xiu.size, neval[1]), order='F')
        xe2 = xe2.reshape((neval[0], etau.size), order='F')
        ye2 = ye2.reshape((neval[0], etau.size), order='F')

        for i in range(xiu.size):
            # loop on xi
            # Getting one eta iso-curve
            plt.plot(xe1[i, :], ye1[i, :], color=edgecolor,
                     alpha=alpha, **kwargs)

        for i in range(etau.size):
            # loop on eta
            # Getting one xi iso-curve
            plt.plot(xe2[:, i], ye2[:, i], color=edgecolor,
                     alpha=alpha, **kwargs)
        plt.plot(Pxm, Pym, color=edgecolor,
                 alpha=alpha, marker='o', linestyle='')
        plt.axis('equal')

    def DegreeElevation(self, new_degree):
        # m.DegreeElevation(np.array([3, 3]))
        new_degree = np.array(new_degree)
        t = new_degree - self.degree
        # to homogeneous coordinates
        nbf_xi, nbf_eta = self.Get_nbf_1d()
        # Degree elevation along the eta direction
        if t[1] != 0:
            coefs = self.ctrlPts.reshape((2*nbf_xi, nbf_eta), order="F")
            self.ctrlPts, self.knotVect[1] = bspdegelev(
                self.degree[1], coefs, self.knotVect[1], t[1])
            nbf_eta = self.ctrlPts.shape[1]
            self.ctrlPts = self.ctrlPts.reshape(
                (2, nbf_xi, nbf_eta), order="F")
        # Degree elevation along the xi direction
        if t[0] != 0:
            coefs = np.transpose(self.ctrlPts, (0, 2, 1))
            coefs = coefs.reshape((2*nbf_eta, nbf_xi), order="F")
            self.ctrlPts, self.knotVect[0] = bspdegelev(
                self.degree[0], coefs, self.knotVect[0], t[0])
            nbf_xi = self.ctrlPts.shape[1]
            self.ctrlPts = self.ctrlPts.reshape(
                (2, nbf_eta, nbf_xi), order="F")
            self.ctrlPts = np.transpose(self.ctrlPts, (0, 2, 1))

        self.degree = new_degree
        self.n = self.CrtlPts2N()

    def KnotInsertion(self, knots):
        # to homogeneous coordinates
        # Example: m.KnotInsertion([np.array([0.5]), np.array([0.5])])
        nbf_xi, nbf_eta = self.Get_nbf_1d()
        nxi = np.size(knots[0])
        neta = np.size(knots[1])
        # Degree elevate along the eta direction
        if neta != 0:
            coefs = self.ctrlPts.reshape((2*nbf_xi, nbf_eta), order="F")
            self.ctrlPts, self.knotVect[1] = bspkntins(
                self.degree[1], coefs, self.knotVect[1], knots[1])
            nbf_eta = self.ctrlPts.shape[1]
            self.ctrlPts = self.ctrlPts.reshape(
                (2, nbf_xi, nbf_eta), order="F")
        # Degree elevate along the xi direction
        if nxi != 0:
            coefs = np.transpose(self.ctrlPts, (0, 2, 1))
            coefs = coefs.reshape((2*nbf_eta, nbf_xi), order="F")
            self.ctrlPts, self.knotVect[0] = bspkntins(
                self.degree[0], coefs, self.knotVect[0], knots[0])
            nbf_xi = self.ctrlPts.shape[1]
            self.ctrlPts = self.ctrlPts.reshape(
                (2, nbf_eta, nbf_xi), order="F")
            self.ctrlPts = np.transpose(self.ctrlPts, (0, 2, 1))
        self.n = self.CrtlPts2N()

    def Stiffness(self, hooke):
        """ 
        Stiffness Matrix 
        """
        wg = sps.diags(self.wdetJ)
        Bxy = self.dphixdy+self.dphiydx
        K = hooke[0, 0]*self.dphixdx.T.dot(wg.dot(self.dphixdx)) +   \
            hooke[1, 1]*self.dphiydy.T.dot(wg.dot(self.dphiydy)) +   \
            hooke[2, 2]*Bxy.T.dot(wg.dot(Bxy)) + \
            hooke[0, 1]*self.dphixdx.T.dot(wg.dot(self.dphiydy)) +   \
            hooke[0, 2]*self.dphixdx.T.dot(wg.dot(Bxy)) +  \
            hooke[1, 2]*self.dphiydy.T.dot(wg.dot(Bxy)) +  \
            hooke[1, 0]*self.dphiydy.T.dot(wg.dot(self.dphixdx)) +   \
            hooke[2, 0]*Bxy.T.dot(wg.dot(self.dphixdx)) +  \
            hooke[2, 1]*Bxy.T.dot(wg.dot(self.dphiydy))
        return K

    def Laplacian(self):
        wg = sps.diags(self.wdetJ)
        return self.dphixdx.T.dot(wg.dot(self.dphixdx)) + self.dphixdy.T.dot(wg.dot(self.dphixdy)) +\
            self.dphiydx.T.dot(wg.dot(self.dphiydx)) + \
            self.dphiydy.T.dot(wg.dot(self.dphiydy))

    def DoubleLaplacian(self):
        wg = sps.diags(self.wdetJ)
        return 2*self.dphixdxx.T.dot(wg.dot(self.dphixdyy)) +\
            2*self.dphiydxx.T.dot(wg.dot(self.dphiydyy)) +\
            self.dphixdxx.T.dot(wg.dot(self.dphixdxx)) +\
            self.dphixdyy.T.dot(wg.dot(self.dphixdyy)) +\
            self.dphiydxx.T.dot(wg.dot(self.dphiydxx)) +\
            self.dphiydyy.T.dot(wg.dot(self.dphiydyy))

    def GaussIntegration(self, npg=None, P=None):
        """ Gauss integration: build of the global differential operators """
        if npg is None:
            nbg_xi = self.degree[0]+1
            nbg_eta = self.degree[1]+1
        else:
            nbg_xi = npg[0]
            nbg_eta = npg[1]

        Gauss_xi = GaussLegendre(nbg_xi)
        Gauss_eta = GaussLegendre(nbg_eta)
        nbf = self.Get_nbf()

        e_xi = np.unique(self.knotVect[0])
        ne_xi = e_xi.shape[0]-1
        e_eta = np.unique(self.knotVect[1])
        ne_eta = e_eta.shape[0]-1
        xi_min = np.kron(e_xi[:-1], np.ones(nbg_xi))
        eta_min = np.kron(e_eta[:-1], np.ones(nbg_eta))
        xi_g = np.kron(np.ones(ne_xi), Gauss_xi[0])
        eta_g = np.kron(np.ones(ne_eta), Gauss_eta[0])

        """ Measures of elements """
        mes_xi = e_xi[1:] - e_xi[:-1]
        mes_eta = e_eta[1:] - e_eta[:-1]

        mes_xi = np.kron(mes_xi, np.ones(nbg_xi))
        mes_eta = np.kron(mes_eta, np.ones(nbg_eta))

        """ Going from the reference element to the parametric space  """
        xi = xi_min + 0.5*(xi_g+1) * \
            mes_xi     # Aranged gauss points in  xi direction
        # Aranged gauss points in  eta direction
        eta = eta_min + 0.5*(eta_g+1)*mes_eta

        phi_xi, dphi_xi = global_basisfuns(
            self.degree[0], self.knotVect[0], xi)
        phi_eta, dphi_eta = global_basisfuns(
            self.degree[1], self.knotVect[1], eta)

        phi = sps.kron(phi_eta,  phi_xi,  'csc')
        dphidxi = sps.kron(phi_eta,  dphi_xi,  'csc')
        dphideta = sps.kron(dphi_eta,  phi_xi,  'csc')
        self.npg = phi.shape[0]

        wg_xi = np.kron(np.ones(ne_xi), Gauss_xi[1])
        wg_eta = np.kron(np.ones(ne_eta), Gauss_eta[1])

        mes_xi = np.kron(np.ones(eta.shape[0]), mes_xi)
        mes_eta = np.kron(mes_eta, np.ones(xi.shape[0]))

        if P is None:
            P = self.Get_P()

        """ Jacobian elements"""
        dxdxi = dphidxi.dot(P[:, 0])
        dxdeta = dphideta.dot(P[:, 0])
        dydxi = dphidxi.dot(P[:, 1])
        dydeta = dphideta.dot(P[:, 1])
        detJ = dxdxi*dydeta - dydxi*dxdeta
        """ Spatial derivatives """
        dphidx = sps.diags(dydeta/detJ).dot(dphidxi) + \
            sps.diags(-dydxi/detJ).dot(dphideta)
        dphidy = sps.diags(-dxdeta/detJ).dot(dphidxi) + \
            sps.diags(dxdxi/detJ).dot(dphideta)
        """ Integration weights + measures + Jacobian of the transformation """
        self.wdetJ = np.kron(wg_eta, wg_xi)*np.abs(detJ)*mes_xi*mes_eta/4
        zero = sps.csr_matrix((self.npg, nbf))
        self.phi = phi
        self.phix = sps.hstack((phi, zero),  'csc')
        self.phiy = sps.hstack((zero, phi),  'csc')
        self.dphixdx = sps.hstack((dphidx, zero),  'csc')
        self.dphixdy = sps.hstack((dphidy, zero),  'csc')
        self.dphiydx = sps.hstack((zero, dphidx),  'csc')
        self.dphiydy = sps.hstack((zero, dphidy),  'csc')

    def GetApproxElementSize(self, cam=None):
        if cam is None:
            # in physical unit
            u, v = self.n[:, 0], self.n[:, 1]
            m2 = self.Copy()
            m2.GaussIntegration(npg=[1, 1], P=np.c_[u, v])
            n = np.max(np.sqrt(m2.wdetJ))
        else:
            # in pyxel unit (int)
            u, v = cam.P(self.n[:, 0], self.n[:, 1])
            m2 = self.Copy()
            m2.GaussIntegration(npg=[1, 1], P=np.c_[u, v])
            n = int(np.floor(np.max(np.sqrt(m2.wdetJ))))
        return n

    def DICIntegrationFast(self, n=10):
        self.DICIntegration(n)

    def DICIntegration(self, n=10):
        """ DIC integration: build of the global differential operators """
        if hasattr(n, 'rz'):
            # if n is a camera then n is autocomputed
            n = self.GetApproxElementSize(n)
        if type(n) == int:
            n = np.array([n, n], dtype=int)
        n = np.maximum(self.degree + 1, n)
        nbg_xi = n[0]
        nbg_eta = n[1]

        Rect_xi = np.linspace(-1, 1, nbg_xi)
        Weight_xi = 2/n[0] * np.ones(nbg_xi)
        Rect_eta = np.linspace(-1, 1, nbg_eta)
        Weight_eta = 2/n[1] * np.ones(nbg_eta)

        nbf = self.Get_nbf()

        e_xi = np.unique(self.knotVect[0])
        ne_xi = e_xi.shape[0]-1
        e_eta = np.unique(self.knotVect[1])
        ne_eta = e_eta.shape[0]-1
        xi_min = np.kron(e_xi[:-1], np.ones(nbg_xi))
        eta_min = np.kron(e_eta[:-1], np.ones(nbg_eta))
        xi_g = np.kron(np.ones(ne_xi), Rect_xi)
        eta_g = np.kron(np.ones(ne_eta), Rect_eta)

        """ Measures of elements """
        mes_xi = e_xi[1:] - e_xi[:-1]
        mes_eta = e_eta[1:] - e_eta[:-1]

        mes_xi = np.kron(mes_xi, np.ones(nbg_xi))
        mes_eta = np.kron(mes_eta, np.ones(nbg_eta))

        """ Going from the reference element to the parametric space  """
        xi = xi_min + 0.5*(xi_g+1) * \
            mes_xi     # Aranged gauss points in  xi direction
        # Aranged gauss points in  eta direction
        eta = eta_min + 0.5*(eta_g+1)*mes_eta

        phi_xi, dphi_xi = global_basisfuns(
            self.degree[0], self.knotVect[0], xi)
        phi_eta, dphi_eta = global_basisfuns(
            self.degree[1], self.knotVect[1], eta)

        phi = sps.kron(phi_eta,  phi_xi,  'csc')
        dphidxi = sps.kron(phi_eta,  dphi_xi,  'csc')
        dphideta = sps.kron(dphi_eta,  phi_xi,  'csc')
        self.npg = phi.shape[0]

        wg_xi = np.kron(np.ones(ne_xi), Weight_xi)
        wg_eta = np.kron(np.ones(ne_eta), Weight_eta)

        mes_xi = np.kron(np.ones(eta.shape[0]), mes_xi)
        mes_eta = np.kron(mes_eta, np.ones(xi.shape[0]))

        P = self.Get_P()

        """ Jacobian elements"""
        dxdxi = dphidxi.dot(P[:, 0])
        dxdeta = dphideta.dot(P[:, 0])
        dydxi = dphidxi.dot(P[:, 1])
        dydeta = dphideta.dot(P[:, 1])
        detJ = dxdxi*dydeta - dydxi*dxdeta
        """ Spatial derivatives """
        dphidx = sps.diags(dydeta/detJ).dot(dphidxi) + \
            sps.diags(-dydxi/detJ).dot(dphideta)
        dphidy = sps.diags(-dxdeta/detJ).dot(dphidxi) + \
            sps.diags(dxdxi/detJ).dot(dphideta)
        """ Integration weights + measures + Jacobian of the transformation """
        self.wdetJ = np.kron(wg_eta, wg_xi)*np.abs(detJ)*mes_xi*mes_eta/4

        zero = sps.csr_matrix((self.npg, nbf))
        self.phi = phi
        self.phix = sps.hstack((phi, zero),  'csc')
        self.phiy = sps.hstack((zero, phi),  'csc')
        self.dphixdx = sps.hstack((dphidx, zero),  'csc')
        self.dphixdy = sps.hstack((dphidy, zero),  'csc')
        self.dphiydx = sps.hstack((zero, dphidx),  'csc')
        self.dphiydy = sps.hstack((zero, dphidy),  'csc')

        self.pgx = self.phi @ P[:, 0]
        self.pgy = self.phi @ P[:, 1]

    def SetBasisFunctionsAtIntegrationPoints(self):
        phi, dphidx, dphidy = Get2dBasisFunctionsAtPts(
            self.pix, self.piy, self.knotVect[0], self.knotVect[1], self.degree[0], self.degree[1])
        nbf = self.Get_nbf()
        self.wg = sps.diags(self.wg)
        zero = sps.csr_matrix((self.npg, nbf))
        self.phiMatrix = phi
        self.phix = sps.hstack((phi, zero),  'csc')
        self.phiy = sps.hstack((zero, phi),  'csc')
        self.dphixdx = sps.hstack((dphidx, zero),  'csc')
        self.dphixdy = sps.hstack((dphidy, zero),  'csc')
        self.dphiydx = sps.hstack((zero, dphidx),  'csc')
        self.dphiydy = sps.hstack((zero, dphidy),  'csc')

    def ShapeFunctionsAtGridPoints(self, xi, eta):
        """ xi and eta are the 1d points 
        This method computes the basis functions on the mesh-grid point 
        obtained from the 1d vector points xi and eta 
        """

        phi_xi, dphi_xi = global_basisfuns(
            self.degree[0], self.knotVect[0], xi)
        phi_eta, dphi_eta = global_basisfuns(
            self.degree[1], self.knotVect[1], eta)

        phi = sps.kron(phi_eta,  phi_xi,  'csc')
        dphidxi = sps.kron(phi_eta,  dphi_xi,  'csc')
        dphideta = sps.kron(dphi_eta,  phi_xi,  'csc')

        P = self.Get_P()

        """ Jacobian elements"""
        dxdxi = dphidxi.dot(P[:, 0])
        dxdeta = dphideta.dot(P[:, 0])
        dydxi = dphidxi.dot(P[:, 1])
        dydeta = dphideta.dot(P[:, 1])
        detJ = dxdxi*dydeta - dydxi*dxdeta
        """ Spatial derivatives """
        dphidx = sps.diags(dydeta/detJ).dot(dphidxi) + \
            sps.diags(-dydxi/detJ).dot(dphideta)
        dphidy = sps.diags(-dxdeta/detJ).dot(dphidxi) + \
            sps.diags(dxdxi/detJ).dot(dphideta)
        # Univariate basis functions if needed
        # Nxi  = phi_xi
        # Neta = phi_eta
        N = phi
        return N, dphidx, dphidy

    def GetBoundaryIndices(self):
        P = self.Get_P()
        left = np.where(np.abs(P[:, 0] - np.min(P[:, 0])) <= 1.e-12)[0]
        right = np.where(np.abs(P[:, 0] - np.max(P[:, 0])) <= 1.e-12)[0]
        bottom = np.where(np.abs(P[:, 1] - np.min(P[:, 1])) <= 1.e-12)[0]
        top = np.where(np.abs(P[:, 1] - np.max(P[:, 1])) <= 1.e-12)[0]
        sleft = np.size(left)
        stop = np.size(top)
        sbottom = np.size(bottom)
        sright = np.size(right)
        return bottom, top, right, left, sbottom, stop, sright, sleft

    def PlaneWave(self, T):
        V = np.zeros(self.ndof)
        V[self.conn[:, 0]] = np.cos(self.n[:, 1] / T * 2 * np.pi)
        return V

    def RBM(self):
        """
        INFINITESIMAL RIGID BODY MODES

        Returns
        -------
        tx : 1D NUMPY ARRAY
            Give the dof vector corresponding to a unitary rigid body
            translation in direction x.
        ty : 1D NUMPY ARRAY
            Give the dof vector corresponding to a unitary rigid body
            translation in direction y.
        rz : 1D NUMPY ARRAY
            Give the dof vector corresponding to a infinitesimal unitary rigid
            body rotation around direction z.

        """
        if self.dim == 3:
            tx = np.zeros(self.ndof)
            tx[self.conn[:, 0]] = 1
            ty = np.zeros(self.ndof)
            ty[self.conn[:, 1]] = 1
            tz = np.zeros(self.ndof)
            tz[self.conn[:, 2]] = 1
            v = self.n - np.mean(self.n, axis=0)
            amp = np.max(np.linalg.norm(v, axis=1))
            rx = np.zeros(self.ndof)
            rx[self.conn] = np.c_[0*v[:, 0], v[:, 2], -v[:, 1]] / amp
            ry = np.zeros(self.ndof)
            ry[self.conn] = np.c_[-v[:, 2], 0*v[:, 1], v[:, 0]] / amp
            rz = np.zeros(self.ndof)
            rz[self.conn] = np.c_[v[:, 1], -v[:, 0], 0*v[:, 2]] / amp
            return tx, ty, tz, rx, ry, rz
        else:
            tx = np.zeros(self.ndof)
            tx[self.conn[:, 0]] = 1
            ty = np.zeros(self.ndof)
            ty[self.conn[:, 1]] = 1
            v = self.n-np.mean(self.n, axis=0)
            v = np.c_[-v[:, 1], v[:, 0]] / np.max(np.linalg.norm(v, axis=1))
            rz = np.zeros(self.ndof)
            rz[self.conn] = v
            return tx, ty, rz


def GaussLegendre(n):
    #     [nodes,weigths]=GaussLegendre(n)
    #
    # Generates the abscissa and weights for a Gauss-Legendre quadrature.
    # Reference:  Numerical Recipes in Fortran 77, Cornell press.

    # Preallocations.
    xg = np.zeros(n)
    wg = np.zeros(n)
    m = (n+1)/2
    #import pdb; pdb.set_trace()
    for ii in range(int(m)):  # for ii=1:m
        # Initial estimate.
        z = np.cos(np.pi*(ii+1-0.25)/(n+0.5))
        z1 = z+1
        while np.abs(z-z1) > np.finfo(np.float64).eps:
            p1 = 1
            p2 = 0
            for jj in range(n):  # for jj = 1:n
                p3 = p2
                p2 = p1
                # The Legendre polynomial.
                p1 = ((2*jj+1)*z*p2-(jj)*p3)/(jj+1)

            # The L.P. derivative.
            pp = n*(z*p1-p2)/(z**2-1)
            z1 = z
            z = z1-p1/pp

        xg[ii] = -z                                   # Build up the abscissas.
        xg[-1-ii] = z
        # Build up the weights.
        wg[ii] = 2/((1-z**2)*(pp**2))
        wg[-1-ii] = wg[ii]
    return xg, wg


def SplineFromROI(roi, dx, degree=[2, 2]):
    """Build a structured FE mesh and a pyxel.camera object from a region
    of interest (ROI) selected in an image f

    Parameters
    ----------
    roi : numpy.array
        The Region of Interest made using  f.SelectROI(), f being a pyxel.Image
    dx : numpy or python array
        dx  = [dx, dy]: average element size (can be scalar) in pixels
    typel : int
        type of element: {3: 'qua4',2: 'tri3',9: 'tri6',16: 'qua8',10: 'qua9'}

    Returns
    -------
    m : pyxel.Mesh
        The finite element mesh
    cam : pyxel.Camera
        The corresponding camera

    Example:
    -------
    f.SelectROI()  -> select the region with rectangle selector
                      and copy - paste the roi in the python terminal
    m, cam = px.MeshFromROI(roi, [20, 20], 3)
    """
    dbox = roi[1] - roi[0]
    NE = (dbox / dx).astype(int)
    NE = np.max(np.c_[NE, np.ones(2, dtype=int)], axis=1)
    m = Rectangle(roi, NE, degree)
    m.n[:, 1] *= -1
    m.ctrlPts = m.N2CrtlPts()
    p = np.array([1., 0., 0., 0.])
    cam = Camera(p)
    return m, cam


def Rectangle(roi, n_elems, degree):
    xmin, ymin, xmax, ymax = roi.ravel()
    # Parametric space properties
    p = 1
    q = 1
    Xi = np.concatenate(
        (np.repeat(0, p+1), np.repeat(1, p+1)))*(xmax-xmin) + xmin
    Eta = np.concatenate(
        (np.repeat(0, q+1), np.repeat(1, q+1)))*(ymax-ymin) + ymin
    # Control points for a recangular plate
    x = np.array([[xmin, xmin],
                  [xmax, xmax]])
    y = np.array([[ymin, ymax],
                  [ymin, ymax]])
    ctrlPts = np.array([x, y])
    knot_vector = [Xi, Eta]
    m = BSplinePatch(ctrlPts, np.array([p, q]), knot_vector)
    # Degree elevation
    m.DegreeElevation(degree)
    # Knot refinement
    ubar = [None]*2
    ubar[0] = 1/n_elems[0] * np.arange(1, n_elems[0]) * (xmax-xmin) + xmin
    ubar[1] = 1/n_elems[1] * np.arange(1, n_elems[1]) * (ymax-ymin) + ymin
    m.KnotInsertion(ubar)
    # Building connectivity
    m.Connectivity()
    return m




import numpy as np
import scipy.special as spe
import scipy.sparse as sps


def KnotInsertionCoefficients(U, u, p):
    """ Returns the knot insertion coefficients """
    k = findKnotSpan(u, U, p)
    l = len(U)-p-1
    alpha = np.zeros(l+1)
    alpha[:k-p+1] = 1
    alpha[k-p+1:k+1] = (u-U[k-p+1:k+1])/(U[k+1:k+p+1]-U[k-p+1:k+1])
    alpha[k+1:] = 0
    return alpha, k


def KnotInsertionOperatorOneKnot(U, u, p):
    """ Consttucts the uni-variate knot refinement operator for one knot insertion """
    alpha, k = KnotInsertionCoefficients(U, u, p)
    nbf = len(alpha)-1
    indexI = np.kron(np.arange(nbf), np.ones(2))
    indexJ = np.c_[np.arange(nbf), np.arange(1, nbf+1)].ravel()
    nnz_values = np.c_[alpha[:-1], 1-alpha[1:]].ravel()
    M = sps.csc_matrix((nnz_values, (indexI, indexJ)), shape=(nbf, nbf+1))
    return M, k


def KnotInsertionOperatorMultipleKnots(U, um, p):
    """ Constructs the uni-variate knot refinement operator for multiple knot insertion"""
    if len(um) == 0:
        raise ValueError('Must insert at least one knot')
    Uc = np.copy(U)
    M, k = KnotInsertionOperatorOneKnot(Uc, um[0], p)
    Uc = np.insert(Uc, k+1, um[0])
    if len(um > 1):
        for i in range(1, len(um)):
            Mo, k = KnotInsertionOperatorOneKnot(Uc, um[i], p)
            M = M.dot(Mo)
            Uc = np.insert(Uc, k+1, um[i])
    return M


def KnotInsertionOperator2d(p, U, um, V, vm):
    """
    returns the 2d refinement opertor  by the tensor product operation
    Coarse knot vectors in Xi and Eta directions: cXi,cEta 
    Fine knot vectors in Xi and Eta directions: rXi,rEta
    """
    Cu = KnotInsertionOperatorMultipleKnots(U, um, p)
    Cv = KnotInsertionOperatorMultipleKnots(V, vm, p)

    C = sps.kron(Cv, Cu)
    return C


def bezier_extraction_nurbs_1d(U, m, p):
    C = []
    a = p+1
    b = a+1
    nb = 1
    C.append(np.identity(p+1))
    while b < m:
        C.append(np.identity(p+1))  # Initialize the next extraction operator.
        i = b
        # Count multiplicity of the knot at location b.
        while b < m and U[b] == U[b-1]:
            b = b+1
        mult = b-i+1
        if mult < p:
            #Use (10) to compute the alphas.
            numer = U[b-1]-U[a-1]
            alphas = [0 for x in range(p)]
            for j in range(p, mult, -1):
                alphas[j-mult-1] = numer / (U[a+j-1]-U[a-1])
            r = p-mult
            #Update the matrix coefficients for r new knots
            for j in range(1, r+1):
                save = r-j+1
                s = mult+j
                for k in range(p+1, s, -1):
                    alpha = alphas[k-s-1]
                    #The following line corresponds to (9).
                    C[-2][:, k-1] = alpha*C[-2][:, k-1] + \
                        (1.0-alpha)*C[-2][:, k-2]
                if b < m:
                    #Update overlapping coefficients of the next operator.
                    C[-1][save-1:j+save, save-1] = C[-2][p-j:p+1, p]
            nb = nb + 1  # Finished with the current operator.
            if b < m:
                #Update indices for the next operator.
                a = b
                b = b+1
    C.pop()
    return C, nb


# def Oslo1(p, coarsekn, finekn, rf ):
#     """ Knot insertion coefficients
#     using Oslo algorithm
#     Taken from Article Multi level Bezier extraction
#     for hierarchical local refinement of IGA """
#     # rf in range(m-p-1)
#     cf = findKnotSpan(finekn[rf],coarsekn,p)
#     b = 1
#     for k in range(1,p+1):
#         t1 = coarsekn[cf-k+1:cf+1]
#         t2 = coarsekn[cf+1:cf+k+1]
#         x = finekn[rf+k]
#         w = (x-t1)/(t2-t1)
#         b = np.r_[(1-w)*b,0] + np.r_[0,w*b]
#     return b


# def KnotInsertionOperator(p,coarsekn,finekn):
#     m = len(finekn)
#     rf = 0
#     C = Oslo1(p,coarsekn,finekn,rf)
#     for rf in range(1,m-p-1):
#         b = Oslo1(p,coarsekn,finekn,rf)
#         C = np.c_[C,b]
#     return C
# def KnotInsertionOperator2d(p,cXi,cEta,fXi,fEta):
#     """
#     returns the 2d refinement opertor
#     Coarse knot vectors in Xi and Eta directions: cXi,cEta
#     Fine knot vectors in Xi and Eta directions: rXi,rEta
#     """
#     Cxi  = KnotInsertionOperator(p,cXi,fXi)
#     Ceta = KnotInsertionOperator(p,cEta,fEta)

#     C = np.kron(Ceta,Cxi)
#     return C

#def MultipleKnotInsertionsOperator(p,coarsekn,knots):
#    coarseTemp = coarsekn.copy()
#    k = knots[0]
#    cf = findKnotSpan(k,coarseTemp,p)
#    finekn = np.insert(coarseTemp, cf+1, k)
#    C = OneKnotInsertionOperator(p,coarseTemp,finekn)
#    coarseTemp = finekn
#    print(C)
#    for i in range(len(knots)-1):
#            k = knots[i+1]
#            cf = findKnotSpan(k,coarseTemp,p)
#            finekn = np.insert(coarseTemp, cf+1, k)
#            print(k,cf,finekn, coarseTemp)
#            print(OneKnotInsertionOperator(p,coarseTemp,finekn))
#            C = C @ OneKnotInsertionOperator(p,coarseTemp,finekn)
#            coarseTemp = finekn
#    return C


def BernsteinRef(u, p):
    """ Evaluates Bernsein basis functions defined on [-1,1]"""
    i = np.arange(p+1)
    ur = np.repeat(u, p+1).reshape((len(u), p+1))
    N = ((1-ur)**(p-i)*(1+ur)**i)*spe.comb(p, i)/(2**p)
    N1 = ((1-ur)**(p-1-(i-1))*(1+ur)**(i-1))*spe.comb(p-1, i-1)/(2**(p-1))
    N2 = ((1-ur)**(p-1-(i))*(1+ur)**(i))*spe.comb(p-1, i)/(2**(p-1))
    dN = 0.5*p*(N1-N2)
    return N, dN


def nubsconnect(p, n):
    """
    It returns the connectivity table INE for NURBS 
    nel : element index in the total connectivity 
    nen : number of functions per support 
    p : degree of functions 
    n : number of functions = number of control points 
    
    fonction qui donne la table de connectivite INE pour un NURBS
    Numerotation en arriere pour trouver directement NURBS coordinate
    nel nombre d'element total
    nen nombre de fonctions par support
    p degre des fnc
    n nombre de fonctions (nbre de points de controle) """
    nel = n-p
    nen = p+1
    IEN = np.zeros((nen, nel), dtype='int64')
    for i in range(nen):
        for j in range(nel):
            IEN[i, j] = p+j-i
    return IEN


def connect_2D(n_xi, n_eta, p, q):
    """
    Connectivity between 2D elements and control points
    Each parametric element is supported by (p+1)*(q+1) basis function
    """
    nbf_elem = (p+1)*(q+1)
    n_elems = n_xi*n_eta
    IEN = np.zeros((n_elems, nbf_elem))
    k = 0
    for j in range(n_eta):
        for i in range(n_xi):
            t = 0
            # Local functions numbers support (p+1)*(q+1)
            for s in range(q+1):
                for r in range(p+1):
                    IEN[k, t] = (n_xi+p)*(j+s)+i+r
                    t += 1
            k += 1
    return IEN


#%%
def findKnotSpan(u, U, p):
    """
    Finds the knots space of a given knot parameter u in
    the knot vector U corresponding to the degree p 
    """
    m = np.size(U)
    if u == U[m-p-1]:
        k = m-p-2
    else:
        k = np.max(np.where(u >= U))
    return k


def findspan(n, p, u, U):
    return findKnotSpan(u, U, p)


def findspanUniformKnotVector(U, deg, l, u):
    if u == U[len(U)-deg-1]:
        return len(U)-deg-2
    return int(np.floor((u-U[0])/l))+deg


#%% Semble Ok. Bien penser à mettre les knots à ajouter sous forme d'un vecteur np.array (même s'il n'y en a qu'un)
def bspkntins(d, c, k, u):
    ''' Function Name: 
    #
    #   bspkntins - Insert knots into a univariate B-Spline. 
    #  
    # Calling Sequence: 
    #  
    #   [ic,ik] = bspkntins(d,c,k,u) 
    #  
    # Parameters: 
    #  
    #   d	: Degree of the B-Spline. 
    #  
    #   c	: Control points, matrix of size (dim,nc). 
    #  
    #   k	: Knot sequence, row vector of size nk. 
    #  
    #   u	: Row vector of knots to be inserted, size nu 
    #  
    #   ic	: Control points of the new B-Spline, of size (dim,nc+nu) 
    #  
    #   ik	: Knot vector of the new B-Spline, of size (nk+nu) 
    #  
    # Description: 
    #  
    #   Insert knots into a univariate B-Spline. This function provides an 
    #   interface to a toolbox 'C' routine. '''
    mc, nc = c.shape
    nu = len(u)
    nk = len(k)
    #
    # int bspkntins(int d, double *c, int mc, int nc, double *k, int nk,
    #               double *u, int nu, double *ic, double *ik)
    # {
    #   int ierr = 0;
    #   int a, b, r, l, i, j, m, n, s, q, ind;
    #   double alfa;
    #
    #   double **ctrl  = vec2mat(c, mc, nc);
    ic = np.zeros((mc, nc+nu))  # double **ictrl = vec2mat(ic, mc, nc+nu);
    ik = np.zeros(nk+nu)
    #
    n = c.shape[1] - 1  # n = nc - 1;
    r = len(u) - 1  # r = nu - 1;
    #
    m = n + d + 1  # m = n + d + 1;
    a = findspan(n, d, u[0], k)  # a = findspan(n, d, u[0], k);
    b = findspan(n, d, u[r], k)  # b = findspan(n, d, u[r], k);
    b += 1  # ++b;
    #
    for q in range(mc):  # for (q = 0; q < mc; q++)  {
        for j in range(a-d+1):
            # for (j = 0; j <= a-d; j++) ictrl[j][q] = ctrl[j][q];
            ic[q, j] = c[q, j]
        for j in range(b-1, n+1):
            # for (j = b-1; j <= n; j++) ictrl[j+r+1][q] = ctrl[j][q];
            ic[q, j+r+1] = c[q, j]
            #   }

    for j in range(a+1):
        ik[j] = k[j]  # for (j = 0; j <= a; j++)   ik[j] = k[j];
    for j in range(b+d, m+1):
        ik[j+r+1] = k[j]  # for (j = b+d; j <= m; j++) ik[j+r+1] = k[j];
        #
    i = b + d - 1  # i = b + d - 1;
    s = b + d + r  # s = b + d + r;

    for j in range(r, -1, -1):  # for (j = r; j >= 0; j--) {
        while (u[j] <= k[i] and i > a):  # while (u[j] <= k[i] && i > a) {
            for q in range(mc):  # for (q = 0; q < mc; q++)
                ic[q, s-d-1] = c[q, i-d-1]  # ictrl[s-d-1][q] = ctrl[i-d-1][q];

            ik[s] = k[i]  # ik[s] = k[i];
            s -= 1  # --s;
            i -= 1  # --i;
            #     }

        for q in range(mc):  # for (q = 0; q < mc; q++)
            ic[q, s-d-1] = ic[q, s-d]  # ictrl[s-d-1][q] = ictrl[s-d][q];

        for l in range(1, d+1):  # for (l = 1; l <= d; l++)  {
            ind = s - d + l  # ind = s - d + l;
            alfa = ik[s+l] - u[j]  # alfa = ik[s+l] - u[j];
            if abs(alfa) == 0:  # if (fabs(alfa) == 0.0)
                for q in range(mc):  # for (q = 0; q < mc; q++)
                    # ictrl[ind-1][q] = ictrl[ind][q];
                    ic[q, ind-1] = ic[q, ind]
            else:  # else  {
                # alfa /= (ik[s+l] - k[i-d+l]);
                alfa = alfa/(ik[s+l] - k[i-d+l])
                for q in range(mc):  # for (q = 0; q < mc; q++)
                    tmp = (1.-alfa)*ic[q, ind]
                    # ictrl[ind-1][q] = alfa*ictrl[ind-1][q]+(1.0-alfa)*ictrl[ind][q];
                    ic[q, ind-1] = alfa*ic[q, ind-1] + tmp
                    #       }
                    #     }
        #
        ik[s] = u[j]  # ik[s] = u[j];
        s -= 1  # --s;
        #   }
        #
        #   freevec2mat(ctrl);
        #   freevec2mat(ictrl);
        #
        #   return ierr;
        # }
    return ic, ik
#%%


def bincoeff(n, k):
    #  Computes the binomial coefficient.
    #
    #      ( n )      n!
    #      (   ) = --------
    #      ( k )   k!(n-k)!
    #
    #  b = bincoeff(n,k)
    #
    #  Algorithm from 'Numerical Recipes in C, 2nd Edition' pg215.

    # double bincoeff(int n, int k)
    # {
    # return floor(0.5+exp(factln(n)-factln(k)-factln(n-k)))
    b = np.floor(0.5+np.exp(factln(n)-factln(k)-factln(n-k)))

    return b


def factln(n):
    # computes ln(n!)
    if n <= 1:
        f = 0
        return f
    f = spe.gammaln(n+1)  # log(factorial(n));</pre>

    return f
#%%


def bspdegelev(d, c, k, t):
    ''' 
    # Function Name:  
    #   bspdegevel - Degree elevate a univariate B-Spline. 
    # Calling Sequence: 
    #   [ic,ik] = bspdegelev(d,c,k,t) 
    # Parameters: 
    #   d	: Degree of the B-Spline. 
    #   c	: Control points, matrix of size (dim,nc). 
    #   k	: Knot sequence, row vector of size nk. 
    #   t	: Raise the B-Spline degree t times. 
    #   ic	: Control points of the new B-Spline. 
    #   ik	: Knot vector of the new B-Spline. 
    # Description: 
    #   Degree elevate a univariate B-Spline. This function provides an 
    #   interface to a toolbox 'C' routine. 
    '''
    mc, nc = c.shape
    #
    # int bspdegelev(int d, double *c, int mc, int nc, double *k, int nk,
    #                int t, int *nh, double *ic, double *ik)
    # {
    #   int row,col
    #
    #   int ierr = 0;
    #   int i, j, q, s, m, ph, ph2, mpi, mh, r, a, b, cind, oldr, mul;
    #   int n, lbz, rbz, save, tr, kj, first, kind, last, bet, ii;
    #   double inv, ua, ub, numer, den, alf, gam;
    #   double **bezalfs, **bpts, **ebpts, **Nextbpts, *alfs;
    #
    #init ic                                                      #   double **ctrl  = vec2mat(c, mc, nc);
    # double **ictrl = vec2mat(ic, mc, nc*(t+1));
    ic = np.zeros((mc, nc*(t+1)))
    ik = np.zeros((t+1)*k.shape[0])
    #
    n = nc - 1  # n = nc - 1;
    #
    bezalfs = np.zeros((d+1, d+t+1))  # bezalfs = matrix(d+1,d+t+1);
    bpts = np.zeros((mc, d+1))  # bpts = matrix(mc,d+1);
    ebpts = np.zeros((mc, d+t+1))  # ebpts = matrix(mc,d+t+1);
    Nextbpts = np.zeros((mc, d+1))  # Nextbpts = matrix(mc,d+1);
    alfs = np.zeros((d, 1))  # alfs = (double *) mxMalloc(d*sizeof(double));
    #
    m = n + d + 1  # m = n + d + 1;
    ph = d + t  # ph = d + t;
    ph2 = int(ph/2)  # ph2 = ph / 2;
    #
    #   // compute bezier degree elevation coefficeients
    bezalfs[0, 0] = 1.  # bezalfs[0][0] = bezalfs[ph][d] = 1.0;
    bezalfs[d, ph] = 1.                                   #

    # 1:ph2                                               #   for (i = 1; i <= ph2; i++) {
    for i in np.arange(1, ph2+1):
        inv = 1/bincoeff(ph, i)  # inv = 1.0 / bincoeff(ph,i);
        mpi = min(d, i)  # mpi = min(d,i);
        #
        # max(0,i-t):mpi      #     for (j = max(0,i-t); j <= mpi; j++)
        for j in np.arange(max(0, i-t), mpi+1):
            # bezalfs[i][j] = inv * bincoeff(d,j) * bincoeff(t,i-j);
            bezalfs[j, i] = inv*bincoeff(d, j)*bincoeff(t, i-j)
            #
    # ph2+1:ph-1                                          #   for (i = ph2+1; i <= ph-1; i++) {
    for i in np.arange(ph2+1, ph):
        mpi = min(d, i)  # mpi = min(d, i);
        # max(0,i-t):mpi                                   #     for (j = max(0,i-t); j <= mpi; j++)
        for j in np.arange(max(0, i-t), mpi+1):
            # bezalfs[i][j] = bezalfs[ph-i][d-j];
            bezalfs[j, i] = bezalfs[d-j, ph-i]
            #
    mh = ph  # mh = ph;
    kind = ph+1  # kind = ph+1;
    r = -1  # r = -1;
    a = d  # a = d;
    b = d+1  # b = d+1;
    cind = 1  # cind = 1;
    ua = k[0]  # ua = k[0];
    #
    # 0:mc-1                                             #   for (ii = 0; ii < mc; ii++)
    for ii in range(mc):
        ic[ii, 0] = c[ii, 0]  # ictrl[0][ii] = ctrl[0][ii];
    # 0:ph                                                #   for (i = 0; i <= ph; i++)
    for i in range(ph+1):
        ik[i] = ua  # ik[i] = ua;
        #   // initialise first bezier seg
    # 0:d                                                 #   for (i = 0; i <= d; i++)
    for i in range(d+1):
        # 0:mc-1                                          #     for (ii = 0; ii < mc; ii++)
        for ii in range(mc):
            bpts[ii, i] = c[ii, i]  # bpts[i][ii] = ctrl[i][ii];
            #   // big loop thru knot vector
    while b < m:  # while (b < m)  {
        i = b  # i = b;
        while b < m and k[b] == k[b+1]:  # while (b < m && k[b] == k[b+1])
            b = b + 1  # b++;
        mul = b - i + 1  # mul = b - i + 1;
        mh += mul + t  # mh += mul + t;
        ub = k[b]  # ub = k[b];
        oldr = r  # oldr = r;
        r = d - mul  # r = d - mul;
        #
        #     // insert knot u(b) r times
        if oldr > 0:  # if (oldr > 0)
            lbz = np.floor((oldr+2)/2)  # lbz = (oldr+2) / 2;
        else:  # else
            lbz = 1  # lbz = 1;

        if r > 0:  # if (r > 0)
            rbz = ph - np.floor((r+1)/2)  # rbz = ph - (r+1)/2;
        else:  # else
            rbz = ph  # rbz = ph;

        if r > 0:  # if (r > 0) {
            #       // insert knot to get bezier segment
            numer = ub - ua  # numer = ub - ua;
            # d:-1:mul+1                                    #       for (q = d; q > mul; q--)
            for q in np.arange(d, mul, -1):
                # alfs[q-mul-1] = numer / (k[a+q]-ua);
                alfs[q-mul-1] = numer / (k[a+q]-ua)

            # 1:r                                           #       for (j = 1; j <= r; j++)  {
            for j in np.arange(1, r+1):
                save = r - j  # save = r - j;
                s = mul + j  # s = mul + j;
                #
                # d:-1:s                                     #         for (q = d; q >= s; q--)
                for q in np.arange(d, s-1, -1):
                    # 0:mc-1                                 #           for (ii = 0; ii < mc; ii++)
                    for ii in range(mc):
                        tmp1 = alfs[q-s]*bpts[ii, q]
                        tmp2 = (1-alfs(q-s))*bpts(ii, q-1)
                        # bpts[q][ii] = alfs[q-s]*bpts[q][ii]+(1.0-alfs[q-s])*bpts[q-1][ii];
                        bpts[ii, q] = tmp1 + tmp2

                # 0:mc-1                                    #         for (ii = 0; ii < mc; ii++)
                for ii in range(mc):
                    # Nextbpts[save][ii] = bpts[d][ii];
                    Nextbpts[ii, save] = bpts[ii, d]
                    #     // end of insert knot
                    #
                    #     // degree elevate bezier
        # lbz:ph                                           #     for (i = lbz; i <= ph; i++)  {
        for i in np.arange(lbz, ph+1):
            # 0:mc-1                                       #       for (ii = 0; ii < mc; ii++)
            for ii in range(mc):
                ebpts[ii, i] = 0  # ebpts[i][ii] = 0.0;
            mpi = min(d, i)  # mpi = min(d, i);
            # max(0,i-t):mpi                                #       for (j = max(0,i-t); j <= mpi; j++)
            for j in np.arange(max(0, i-t), mpi+1):
                # 0:mc-1                                    #         for (ii = 0; ii < mc; ii++)
                for ii in range(mc):
                    tmp1 = ebpts[ii, i]
                    tmp2 = bezalfs[j, i]*bpts[ii, j]
                    # ebpts[i][ii] = ebpts[i][ii] + bezalfs[i][j]*bpts[j][ii];
                    ebpts[ii, i] = tmp1 + tmp2
                    #     // end of degree elevating bezier
                    #
        if oldr > 1:  # if (oldr > 1)  {
            #       // must remove knot u=k[a] oldr times
            first = kind - 2  # first = kind - 2;
            last = kind  # last = kind;
            den = ub - ua  # den = ub - ua;
            # bet = (ub-ik[kind-1]) / den;
            bet = np.floor((ub-ik[kind-1]) / den)
            #
            #       // knot removal loop
            # 1:oldr-1                                     #       for (tr = 1; tr < oldr; tr++)  {
            for tr in np.arange(1, oldr):
                i = first  # i = first;
                j = last  # j = last;
                kj = j - kind + 1  # kj = j - kind + 1;
                while j-i > tr:  # while (j - i > tr)  {
                    #           // loop and compute the new control points
                    #           // for one removal step
                    if i < cind:  # if (i < cind)  {
                        # alf = (ub-ik[i])/(ua-ik[i]);
                        alf = (ub-ik[i])/(ua-ik[i])
                        # 0:mc-1                              #             for (ii = 0; ii < mc; ii++)
                        for ii in range(mc):
                            tmp1 = alf*ic[ii, i]
                            tmp2 = (1-alf)*ic[ii, i-1]
                            # ictrl[i][ii] = alf * ictrl[i][ii] + (1.0-alf) * ictrl[i-1][ii];
                            ic[ii, i] = tmp1 + tmp2
                    if j >= lbz:  # if (j >= lbz)  {
                        if j-tr <= kind-ph+oldr:  # if (j-tr <= kind-ph+oldr) {
                            # gam = (ub-ik[j-tr]) / den;
                            gam = (ub-ik[j-tr]) / den
                            # 0:mc-1                           #               for (ii = 0; ii < mc; ii++)
                            for ii in range(mc):
                                tmp1 = gam*ebpts[ii, kj]
                                tmp2 = (1-gam)*ebpts[ii, kj+1]
                                # ebpts[kj][ii] = gam*ebpts[kj][ii] + (1.0-gam)*ebpts[kj+1][ii];
                                ebpts[ii, kj] = tmp1 + tmp2
                        else:  # else  {
                            # 0:mc-1                           #               for (ii = 0; ii < mc; ii++)
                            for ii in range(mc):
                                tmp1 = bet*ebpts[ii, kj]
                                tmp2 = (1-bet)*ebpts[ii, kj+1]
                                # ebpts[kj][ii] = bet*ebpts[kj][ii] + (1.0-bet)*ebpts[kj+1][ii];
                                ebpts[ii, kj] = tmp1 + tmp2
                    i += 1  # i++;
                    j -= 1  # j--;
                    kj -= 1  # kj--;
                    #
                first -= 1  # first--;
                last += 1  # last++;
                #     // end of removing knot n=k[a]
                #
                #     // load the knot ua
        if a != d:  # if (a != d)
            # 0:ph-oldr-1                                   #       for (i = 0; i < ph-oldr; i++)  {
            for i in range(ph-oldr):
                ik[kind] = ua  # ik[kind] = ua;
                kind += 1  # kind++;
                #
                #     // load ctrl pts into ic
        # lbz:rbz                                       #     for (j = lbz; j <= rbz; j++)  {
        for j in np.arange(lbz, rbz+1):
            # 0:mc-1                                    #       for (ii = 0; ii < mc; ii++)
            for ii in range(mc):
                ic[ii, cind] = ebpts[ii, j]  # ictrl[cind][ii] = ebpts[j][ii];
            cind += 1  # cind++;
            #
        if b < m:  # if (b < m)  {
            #       // setup for next pass thru loop
            # 0:r-1                                      #       for (j = 0; j < r; j++)
            for j in range(r):
                # 0:mc-1                                 #         for (ii = 0; ii < mc; ii++)
                for ii in range(mc):
                    # bpts[j][ii] = Nextbpts[j][ii];
                    bpts[ii, j] = Nextbpts[ii, j]
            # r:d                                        #       for (j = r; j <= d; j++)
            for j in np.arange(r, d+1):
                # 0:mc-1                                 #         for (ii = 0; ii < mc; ii++)
                for ii in range(mc):
                    # bpts[j][ii] = ctrl[b-d+j][ii];
                    bpts[ii, j] = c[ii, b-d+j]
            a = b  # a = b;
            b += 1  # b++;
            ua = ub  # ua = ub;
            #     }
        else:  # else
            #       // end knot
            # 0:ph                                       #       for (i = 0; i <= ph; i++)
            for i in range(ph+1):
                ik[kind+i] = ub  # ik[kind+i] = ub;
    # End big while loop                                      #   // end while loop
                #
                #   *nh = mh - ph - 1;
                #
                #   freevec2mat(ctrl);
                #   freevec2mat(ictrl);
                #   freematrix(bezalfs);
                #   freematrix(bpts);
                #   freematrix(ebpts);
                #   freematrix(Nextbpts);
                #   mxFree(alfs);
                #
                #   return(ierr);
                # }

    # ajout dû au fait qu'on a initialisé trop grand (car difficile d'estimer la taille de ic et ik avant, dépend entre autres de la multiplicité des knots)
    # on enleve les 0 à la fin du knot vector ik
    ik = np.trim_zeros(ik, 'b')

    # on tronque la matrice des points de contrôle où il faut (revient à enlever les 0, mais si la courbe finit avec un point en (0,0), on n'enlève pas celui-là)
    n = len(ik)-(d+t)-1
    ic = ic[:, 0:n]

    return ic, ik


def derbasisfuns(i, pl, U, nders, u):
    """ 
    i : knot span of u 
    p : degree 
    u : parameter on which we want to evaluate the function 
    nders: number of derivatives 
    U : knot vector 
    
    # i span de u 
    # pl = degrés de la nurbs
    # u = endroit ou l'on veut la fonction
    # nders = numéro de la dérivée désirée
    # U = vecteur de noeud de la fonction """

#    import pdb; pdb.set_trace()
    u_knotl = U.copy()
    left = np.zeros((pl+1))
    right = np.zeros((pl+1))
    ndu = np.zeros((pl+1, pl+1))
    ders = np.zeros((nders+1, pl+1))
    ndu[0, 0] = 1
    for j in range(pl):  # 1:pl
        left[j+1] = u - u_knotl[i-j]  # rq Ali : i-j au lieu de i-j-1
        right[j+1] = u_knotl[i+j+1] - u  # rq : i+j+1 au lieu de i+j
        saved = 0
        for r in range(j+1):  # 0:j-1
            ndu[j+1, r] = right[r+1] + left[j-r+1]
            temp = ndu[r, j]/ndu[j+1, r]
            ndu[r, j+1] = saved + right[r+1]*temp
            saved = left[j-r+1]*temp
        ndu[j+1, j+1] = saved
#    print('checkpoint1 : '+str(ndu))

        # load basis functions
    for j in range(pl+1):  # 0:pl
        ders[0, j] = ndu[j, pl]
#    print('checkpoint2 : '+str(ders))

        # compute derivatives
    for r in range(pl+1):  # 0:pl              # loop over function index
        s1 = 0
        s2 = 1                # alternate rows in array a
        a = np.zeros((nders+1, nders+1))
        a[0, 0] = 1
        # loop to compute kth derivative
        for k in range(nders):   # 1:nders
            d = 0
            rk = r-(k+1)
            pk = pl-(k+1)
            if (r >= (k+1)):
                a[s2, 0] = a[s1, 0]/ndu[pk+1, rk]
                d = a[s2, 0]*ndu[rk, pk]
            if (rk >= -1):
                j1 = 1
            else:
                j1 = -rk
            if ((r-1) <= pk):
                j2 = k
            else:
                j2 = pl-r
            for j in np.arange(j1, j2+0.1):  # j1:j2
                j = int(j)
                a[s2, j] = (a[s1, j] - a[s1, j-1])/ndu[pk+1, rk+j]
                d = d + a[s2, j]*ndu[rk+j, pk]
            if (r <= pk):
                a[s2, k+1] = -a[s1, k]/ndu[pk+1, r]
                d = d + a[s2, k+1]*ndu[r, pk]
            ders[k+1, r] = d
            j = s1
            s1 = s2
            s2 = j            # switch rows

    #     Multiply through by the correct factors

    r = pl
    for k in range(nders):   # 1:nders
        for j in range(pl+1):   # 0:pl
            ders[k+1, j] = ders[k+1, j]*r
        r = r*(pl-(k+1))

    return ders


def BasisFunc(i, u, p, U):
    """
    Evaluates the non zero basis functions N_{i,p}(u)
    for u living in the knot span [U_i,U_{i+1}[
    From the Nurbs Book ( Les Piegl, Wayne Tiller)
    """
    N = np.zeros(p+1)
    left = np.zeros(p+1)
    right = np.zeros(p+1)
    N[0] = 1.
    for j in range(1, p+1):
        left[j] = u-U[i+1-j]
        right[j] = U[i+j]-u
        saved = 0.
        for r in range(j):
            temp = N[r]/(right[r+1]+left[j-r])
            N[r] = saved+right[r+1]*temp
            saved = left[j-r]*temp
        N[j] = saved
    return N


def EvaluateSpline(x, y, Xi, Eta, p, q, c, nx):
    """ Evaluate spline surface """
    # c is the raveled array of control variables
    se = np.zeros(len(x))
    for k in range(len(x)):
        spanx = findKnotSpan(x[k], Xi, p)
        spany = findKnotSpan(y[k], Eta, q)
        Nx = BasisFunc(spanx, x[k], p, Xi)
        Ny = BasisFunc(spany, y[k], q, Eta)
        s = 0
        for j in range(q+1):
            for i in range(p+1):
                s += Ny[j]*Nx[i]*c[spanx-p+i + (spany-q+j)*nx]
        se[k] = s
    return se


def OneBasisFun(i, u, p, UU):
    """
    Computes the basis function N_{i,p} on the point u 
    Warning: Works only for open knot vectors of type 
    U ={0,...,0,  ....  ,  1,...1} where the first 
    and last knots have a multiplicity equal to p+1 
    From the Nurbs Book ( Les Piegl, Wayne Tiller)
    """
    U = np.copy(UU)
    m = U.shape[0]-1

    # Verifying the local support proprety of the Bsplines
    if (u < U[i] or u > U[i+p+1]):
        Nip = 0.
        return Nip
    # Special case evaluation on the first and last knots
    if (i == 0 and u == U[0]) or (i == m-p-1 and u == U[m]):
        Nip = 1
        return Nip

    N = np.zeros(p+1)
    # Initialize zeroth-degree functions
    for j in range(p+1):
        if u >= U[i+j] and u < U[i+j+1]:
            N[j] = 1.
        else:
            N[j] = 0.
    for k in range(1, p+1):
        if N[0] == 0.:
            saved = 0.
        else:
            saved = ((u-U[i])*N[0])/(U[i+k]-U[i])
        for j in range(p-k+1):
            Uleft = U[i+j+1]
            Uright = U[i+j+k+1]
            if N[j+1] == 0.:
                N[j] = saved
                saved = 0.
            else:
                temp = N[j+1]/(Uright-Uleft)
                N[j] = saved+(Uright-u)*temp
                saved = (u-Uleft)*temp
    Nip = N[0]
    return Nip


def derbasisfuncVectorInput(p, U, u, nb_u_values, span, nders):
    """ Returns the basis functions and first derivatives 
    for multiple parameter inputs """
    ders_matrix = np.zeros(((nders+1)*(nb_u_values), p+1))
    for i in range(nb_u_values):
        ders = derbasisfuns(span, p, U, nders, u[i])
        ders_matrix[2*i:2*(i+1), :] = ders
    return ders_matrix[::2, :], ders_matrix[1::2, :]


def secondOrderDerivBasisFuns(p, U, u):
    nb_u_values = len(u)
    nnz_values = nb_u_values*(p+1)
    indexI = np.zeros(nnz_values)
    indexJ = np.zeros(nnz_values)
    ders_values = np.zeros(nnz_values)
    n = len(U) - 1 - p
    k = 0
    for i in range(nb_u_values):
        span = findspan(n, p, u[i], U)
        ders = derbasisfuns(span, p, U, 2, u[i])
        for j in range(p+1):
            indexI[k] = i
            indexJ[k] = span - p + j
            ders_values[k] = ders[2][j]
            k += 1
    phi_second = sps.csc_matrix(
        (ders_values, (indexI, indexJ)), shape=(nb_u_values, n))
    return phi_second


def Get1dBasisFunctionsAtPts(degree, knotVect, pts, der):
    operators = [None]*(der+1)
    nb_pts_values = len(pts)
    nnz_values = nb_pts_values*(degree+1)
    indexI = np.zeros(nnz_values)
    indexJ = np.zeros(nnz_values)
    values = [np.zeros(nnz_values)]*(degree+1)
    n = len(knotVect) - 1 - degree
    k = 0
    for i in range(nb_pts_values):
        span = findspan(n, degree, pts[i], knotVect)
        ders = derbasisfuns(span, degree, knotVect, 1, pts[i])
        for j in range(degree+1):
            indexI[k] = i
            indexJ[k] = span - degree + j
            for d in range(der+1):
                values[d][k] = ders[d][j]
            k += 1
    for d in range(der+1):
        operators[d] = sps.csc_matrix(
            (values[d], (indexI, indexJ)), shape=(nb_pts_values, n))
    return operators


def global_basisfuns(p, U, u):
    """ Returns the global differential operator N and dNdxi
    N and dNdxi are sparse matrices of size (size(u),nbf)
    where nbf is the toral number of basis functions
    """
    nb_u_values = len(u)
    nnz_values = nb_u_values*(p+1)
    indexI = np.zeros(nnz_values)
    indexJ = np.zeros(nnz_values)
    values = np.zeros(nnz_values)
    ders_values = np.zeros(nnz_values)
    n = len(U) - 1 - p
    k = 0
    for i in range(nb_u_values):
        span = findspan(n, p, u[i], U)
        ders = derbasisfuns(span, p, U, 1, u[i])
        for j in range(p+1):
            indexI[k] = i
            indexJ[k] = span - p + j
            values[k] = ders[0][j]
            ders_values[k] = ders[1][j]
            k += 1
    phi = sps.csc_matrix((values, (indexI, indexJ)), shape=(nb_u_values, n))
    dphi = sps.csc_matrix((ders_values, (indexI, indexJ)),
                          shape=(nb_u_values, n))
    return phi, dphi


def global_basisfunsWd(p, U, u):
    """ Returns only the function N without derivatives"""
    nb_u_values = len(u)
    nnz_values = nb_u_values*(p+1)
    indexI = np.zeros(nnz_values)
    indexJ = np.zeros(nnz_values)
    values = np.zeros(nnz_values)
    n = len(U) - 1 - p
    k = 0
    for i in range(nb_u_values):
        span = findspan(n, p, u[i], U)
        N = BasisFunc(span, u[i], p, U)
        for j in range(p+1):
            indexI[k] = i
            indexJ[k] = span - p + j
            values[k] = N[j]
            k += 1
    phi = sps.csc_matrix((values, (indexI, indexJ)), shape=(nb_u_values, n))
    return phi


def Get2dBasisFunctionsAtPts(u, v, U, V, p, q):
    nb_u_values = len(u)
    nnz_values = nb_u_values*(p+1)*(q+1)
    indexI = np.zeros(nnz_values)
    indexJ = np.zeros(nnz_values)
    valuesN = np.zeros(nnz_values)
    valuesdNdu = np.zeros(nnz_values)
    valuesdNdv = np.zeros(nnz_values)

    n = len(U) - 1 - p
    m = len(V) - 1 - q
    nbf = n*m
    l = 0
    for k in range(nb_u_values):
        spanu = findspan(n, p, u[k], U)
        spanv = findspan(m, q, v[k], V)
        Nu = derbasisfuns(spanu, p, U, 1, u[k])
        Nv = derbasisfuns(spanv, q, V, 1, v[k])
        for j in range(q+1):
            for i in range(p+1):
                valuesN[l] = Nv[0][j]*Nu[0][i]
                valuesdNdu[l] = Nv[0][j]*Nu[1][i]
                valuesdNdv[l] = Nv[1][j]*Nu[0][i]
                indexI[l] = k
                indexJ[l] = spanu - p + i + (spanv - q + j)*n
                l = l+1
    phi = sps.csc_matrix((valuesN, (indexI, indexJ)), shape=(nb_u_values, nbf))
    dphidu = sps.csc_matrix(
        (valuesdNdu, (indexI, indexJ)), shape=(nb_u_values, nbf))
    dphidv = sps.csc_matrix(
        (valuesdNdv, (indexI, indexJ)), shape=(nb_u_values, nbf))
    return phi, dphidu, dphidv


def Get3dBasisFunctionsAtPts(x, y, z, Xi, Eta, Zeta, p, q, r):
    nb_x_values = len(x)
    # Number of basis functions that support a point
    nbf_pt = (p+1)*(q+1)*(r+1)
    nnz_values = nb_x_values*nbf_pt
    indexI = np.zeros(nnz_values)
    indexJ = np.zeros(nnz_values)
    valuesN = np.zeros(nnz_values)

    nbf_xi = len(Xi) - 1 - p
    nbf_eta = len(Eta) - 1 - q
    nbf_zeta = len(Zeta) - 1 - r

    nbf = nbf_xi * nbf_eta * nbf_zeta

    index = np.arange(nbf_pt)
    index_i = np.kron(np.ones((r+1)*(q+1)), np.arange(p+1))
    index_j = np.kron(np.ones(r+1), np.kron(np.arange(q+1), np.ones(p+1)))
    index_k = np.kron(np.arange(r+1), np.ones((p+1)*(q+1)))

    #ll=0  # uncomment if unvectorized version is used
    for up in range(nb_x_values):
        # Loop over the unstructured points
        spanx = findKnotSpan(x[up], Xi, p)
        spany = findKnotSpan(y[up], Eta, q)
        spanz = findKnotSpan(z[up], Zeta, r)
        Nxi = BasisFunc(spanx, x[up], p, Xi)
        Neta = BasisFunc(spany, y[up], q, Eta)
        Nzeta = BasisFunc(spanz, z[up], r, Zeta)
        valuesN[index + up*nbf_pt] = np.kron(Nzeta, np.kron(Neta, Nxi))
        indexI[index + up*nbf_pt] = up
        indexJ[index + up*nbf_pt] = (spanx-p+index_i) + (spany -
                                                         q+index_j)*nbf_xi + (spanz-r+index_k)*nbf_xi*nbf_eta
        # Structured grid: i+j*nx+k*nx*ny (arrangement in structured grids)
        # Non vectorized approach
        # for k in range(r+1):
        #     for j in range(q+1):
        #         for i in range(p+1):
        #             valuesN[ll]    = Nw[k]*Nv[j]*Nu[i]
        #             indexI[ll] = up
        #             # Structured grid: i+j*nx+k*nx*ny
        #             indexJ[ll] = spanu-p+i +(spanv-q+j)*n +(spanw-r+k)*n*m
        #             ll = ll +1
    phi = sps.csc_matrix((valuesN,    (indexI, indexJ)),
                         shape=(nb_x_values, nbf))
    return phi


# def Get3dBasisFunctionsAtPts(u,v,w,U,V,W,p,q,r):
#     nb_u_values = len(u);
#     nnz_values = nb_u_values*(p+1)*(q+1)*(r+1)
#     indexI      = np.zeros(nnz_values)
#     indexJ      = np.zeros(nnz_values)
#     valuesN     = np.zeros(nnz_values)
#     valuesdNdu  = np.zeros(nnz_values)
#     valuesdNdv  = np.zeros(nnz_values)
#     valuesdNdw  = np.zeros(nnz_values)

#     n = len(U)-1-p
#     m = len(V)-1-q
#     l = len(W)-1-r

#     nbf = n*m*l
#     ll=0
#     for up in range(nb_u_values):
#         spanu = findspan(n,p,u[up],U)
#         spanv = findspan(m,q,v[up],V)
#         spanw = findspan(l,r,w[up],W)
#         Nu    = derbasisfuns(spanu,p,U,1,u[up])
#         Nv    = derbasisfuns(spanv,q,V,1,v[up])
#         Nw    = derbasisfuns(spanw,r,W,1,w[up])
#         for k in range(r+1):
#             for j in range(q+1):
#                 for i in range(p+1):
#                     valuesN[ll]    = Nw[0][k]*Nv[0][j]*Nu[0][i]
#                     valuesdNdu[ll] = Nw[0][k]*Nv[0][j]*Nu[1][i]
#                     valuesdNdv[ll] = Nw[0][k]*Nv[1][j]*Nu[0][i]
#                     valuesdNdw[ll] = Nw[1][k]*Nv[0][j]*Nu[0][i]
#                     indexI[ll] = up
#                     # Structured grid: i+j*nx+k*nx*ny
#                     indexJ[ll] = spanu-p+i +(spanv-q+j)*n +(spanw-r+k)*n*m
#                     ll = ll +1
#     phi     = sps.csc_matrix(( valuesN,    (indexI,indexJ)), shape = (nb_u_values,nbf) )
#     dphidu  = sps.csc_matrix(( valuesdNdu, (indexI,indexJ)), shape = (nb_u_values,nbf) )
#     dphidv  = sps.csc_matrix(( valuesdNdv, (indexI,indexJ)), shape = (nb_u_values,nbf) )
#     dphidw  = sps.csc_matrix(( valuesdNdw, (indexI,indexJ)), shape = (nb_u_values,nbf) )
#     return phi, dphidu, dphidv, dphidw


def Get3dBasisFunctionsAtPtsWd(u, v, w, U, V, W, p, q, r):
    nb_u_values = len(u)
    nnz_values = nb_u_values*(p+1)*(q+1)*(r+1)
    indexI = np.zeros(nnz_values)
    indexJ = np.zeros(nnz_values)
    valuesN = np.zeros(nnz_values)

    n = len(U)-1-p
    m = len(V)-1-q
    l = len(W)-1-r

    nbf = n*m*l
    ll = 0
    for up in range(nb_u_values):
        spanu = findspan(n, p, u[up], U)
        spanv = findspan(m, q, v[up], V)
        spanw = findspan(l, r, w[up], W)
        Nu = BasisFunc(spanu, u[up], p, U)
        Nv = BasisFunc(spanv, v[up], q, V)
        Nw = BasisFunc(spanw, w[up], r, W)
        for k in range(r+1):
            for j in range(q+1):
                for i in range(p+1):
                    valuesN[ll] = Nw[k]*Nv[j]*Nu[i]
                    indexI[ll] = up
                    # Structured grid: i+j*nx+k*nx*ny
                    indexJ[ll] = spanu-p+i + (spanv-q+j)*n + (spanw-r+k)*n*m
                    ll = ll + 1
    phi = sps.csc_matrix((valuesN,    (indexI, indexJ)),
                         shape=(nb_u_values, nbf))
    return phi


def interpolateLinearly(xmin, xmax, lmin, lmax):
    a = (lmax-lmin)/(xmax-xmin)
    b = (lmin*xmax - lmax*xmin)/(xmax-xmin)
    return a, b


def computeNOELEM3D(nxi, neta, nzeta, p, q, r):
    """ Creates element connectivity for 3D B-splines """
    bf_xi_index = (np.kron(np.ones(nxi), np.arange(p+1)) +
                   np.kron(np.arange(nxi), np.ones(p+1))).reshape((nxi, p+1))
    bf_eta_index = (np.kron(np.ones(neta), np.arange(q+1)) +
                    np.kron(np.arange(neta), np.ones(q+1))).reshape((neta, q+1))
    bf_zeta_index = (np.kron(np.ones(nzeta), np.arange(r+1)) +
                     np.kron(np.arange(nzeta), np.ones(r+1))).reshape((nzeta, r+1))

    noelem = np.kron(np.ones_like(bf_zeta_index),  np.kron(np.ones_like(bf_eta_index), bf_xi_index)) + \
        (nxi+p)*np.kron(np.ones_like(bf_zeta_index), np.kron(bf_eta_index, np.ones_like(bf_xi_index))) + \
        (nxi+p)*(neta+q)*np.kron(bf_zeta_index,
                                 np.kron(np.ones_like(bf_eta_index), np.ones_like(bf_xi_index)))

    return noelem.astype('int32')


# nxi = 4 ; neta = 2
# p = 2   ; q= 1
# bf_xi_index  = (np.kron( np.ones(nxi), np.arange(p+1) ) + np.kron(np.arange(nxi), np.ones(p+1) )).reshape((nxi,p+1))
# bf_eta_index = (np.kron( np.ones(neta), np.arange(q+1) ) + np.kron(np.arange(neta), np.ones(q+1) )).reshape((neta,q+1))
# noelem = np.kron(np.ones_like(bf_eta_index),bf_xi_index) + (nxi+p)*np.kron(bf_eta_index, np.ones_like(bf_xi_index))

# nxi = 2 ; neta = 1 ; nzeta = 3
# p = 2 ; q = 1 ; r = 3
# bf_xi_index   = (np.kron( np.ones(nxi), np.arange(p+1) ) + np.kron(np.arange(nxi), np.ones(p+1) )).reshape((nxi,p+1))
# bf_eta_index  = (np.kron( np.ones(neta), np.arange(q+1) ) + np.kron(np.arange(neta), np.ones(q+1) )).reshape((neta,q+1))
# bf_zeta_index = (np.kron( np.ones(nzeta), np.arange(r+1) ) + np.kron(np.arange(nzeta), np.ones(r+1) )).reshape((nzeta,r+1))

# noelem = np.kron(  np.ones_like(bf_zeta_index) ,  np.kron(np.ones_like(bf_eta_index)  ,bf_xi_index) )  + \
#          (nxi+p)*np.kron(  np.ones_like(bf_zeta_index), np.kron( bf_eta_index, np.ones_like(bf_xi_index) ) )  + \
#          (nxi+p)*(neta+q)*np.kron(  bf_zeta_index, np.kron( np.ones_like(bf_eta_index),np.ones_like(bf_xi_index) ) )


# noelemC = np.zeros((nxi*neta*nzeta,(p+1)*(q+1)*(r+1)))

# ie=0;
# for k in range(nzeta):
#     for j in range(neta):
#         for i in range(nxi):
#             t=0
#             for kk in range(r+1):
#                 for jj in range(q+1):
#                     for ii in range(p+1):
#                         noelemC[ie,t] = i+ii+(j+jj)*(nxi+p)+(k+kk)*(nxi+p)*(neta+q) ;
#                         t+=1
#             ie+=1

# print(np.max(noelem-noelemC))
# print(np.min(noelem-noelemC))
